import numpy as np
import metamathpy.setmm as ms
import matplotlib.pyplot as pt

db = ms.load_pl()
# db = ms.load_all()

# number of essentials for each rule
less = [len(rule.essentials) for rule in db.rules.values()]

# histogram bins
bins = np.arange(max(less)+1)+.5 # uniform buckets
# bins = # log scale buckets

hist, _ = np.histogram(less, bins)

pt.bar(bins[1:]+.5, hist, align='center')
pt.xscale('log')
pt.yscale('log')
pt.show()




"""
pip install blessed
"""
from blessed import Terminal

term = Terminal()

with term.fullscreen(), term.no_line_wrap():
    print(term.home + term.clear + term.move_y(term.height // 2))
    print(term.black_on_darkkhaki(term.center('press any key to continue.')))
    
    with term.cbreak(), term.hidden_cursor():
        inp = term.inkey()
    
    print(term.move_down(2) + 'You pressed ' + term.bold(repr(inp)))

    input("Enter to end...")
    



# run from mmpy with $ python -m src.mmmine.dependencies
import sys
import os
import random
import itertools as it
import pickle as pk
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms

if __name__ == "__main__":

    do_deps = False

    import matplotlib.pyplot as pt
    from matplotlib import rcParams
    rcParams["font.family"] = "serif"

    # db = ms.load_imp()
    # db = ms.load_ni()
    db = ms.load_pl()
    # db = ms.load_all()
    # db.print()

    # build direct dependency graph
    direct_dependencies = {} # rule label -> set of labels in rule's proof
    proof_lengths = {}

    for r, (label, rule) in enumerate(db.rules.items()):
        # only propositions have proofs
        if rule.consequent.tag != "$p": continue

        # get proof labels (assumes proof in compressed format)
        split = rule.consequent.proof.index(")")
        step_labels = rule.consequent.proof[1:split]

        direct_dependencies[label] = set(step_labels)

        # get proof size
        root, steps = mp.verify_proof(db, rule)
        proof_lengths[label] = len(steps)


    # direct dependency histogram
    num_deps = sorted(map(len, direct_dependencies.values()))

    # pt.plot(num_deps)
    # pt.xlabel("Sorted rule index")
    # pt.ylabel("# direct dependencies")
    # pt.show()

    if do_deps:

        # dependents: how many other rules depend on this one, either directly or indirectly
        num_dependents = {label: 0 for label in direct_dependencies}
    
        # recursive helper
        def increment_indirect_dependencies(label):
            if label not in direct_dependencies: return
            for dep in direct_dependencies[label]:
                if dep not in direct_dependencies: continue
                num_dependents[dep] = num_dependents.get(dep, 0) + 1
                increment_indirect_dependencies(dep)
        for label in direct_dependencies:
            increment_indirect_dependencies(label)
    
        # direct dependency histogram
        labels = tuple(num_dependents.keys())
        num_deps = np.array([num_dependents[label] for label in labels])
        plens = np.array([proof_lengths[label] for label in labels])

        with open("dependencies.pkl","wb") as f: pk.dump((num_deps, plens), f)
        
    with open("dependencies.pkl","rb") as f: num_deps, plens = pk.load(f)

    pt.figure(figsize=(5,4))
    pt.scatter(num_deps + 1 + np.random.randn(len(num_deps))*0.1, plens + np.random.randn(len(plens))*0.1, marker='.', color='k')
    pt.xlabel("Number of dependents (+1)")
    pt.ylabel("Proof length")
    pt.xscale("log")
    pt.yscale("log")
    pt.ylim([1,101])
    pt.savefig("dependencies.eps")
    pt.tight_layout()
    pt.show()







# run from mmpy with $ python -m src.mmmine.eliminate_defs
import sys
import os
import random
import itertools as it
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms






from metamathpy.database import parse
from metamathpy.proof import verify_proof

if __name__ == "__main__":

    import os
    fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = parse(fpath)
    print("parsed.")

    # needs outer loop over all proofs in db, and all wrapped in helper function
    rule = db.rules["a1i"]
    _, proof_steps = verify_proof(db, rule)

    examples = []
    for proof_step in proof_steps.values():

        prompt = " ".join(proof_step.conclusion)
        for hyp in rule.essentials:
            prompt += " " + " ".join(hyp.tokens)

        output = " ".join(proof_step.rule.consequent.tokens)
        for hyp in proof_step.rule.essentials:
            output += " " + " ".join(hyp.tokens)

        examples.append( (prompt, output) )

    for n, (prompt, output) in enumerate(examples):
        print(n)
        print(prompt)
        print(output)
        




"""
Run from top-level directory with
$ python -m src.metamathpy.proof
"""
import itertools as it
from metamathpy.unification import substitute, unify

try:
    profile
except NameError:
    profile = lambda x: x


"""
Proof step: represents one step of a proof
conclusion: the symbol string that is concluded by applying an inference rule
rule: the rule object that was applied
dependencies[h]: another ProofStep object, whose conclusion satisfies the rule hypothesis with label h
substitution: the substitution that transforms the rule's consequent and hypotheses into the conclusion and dependencies
disjoint: the set of disjoint variable requirements inherited from dependencies
"""
class ProofStep:
    def __init__(self, conclusion, rule, dependencies=None, substitution=None, disjoint=None):

        # defaults
        if dependencies is None: dependencies = {}
        if substitution is None: substitution = {}
        if disjoint is None: disjoint = set()

        self.conclusion = conclusion # conclusion of the step
        self.rule = rule # justification for the step
        self.dependencies = dependencies # previous steps it relies on
        self.substitution = substitution # substitution that matches them
        self.disjoint = disjoint # disjoint requirements inherited from them

        # cache normal proofs after first construction
        self._normal_proof = None

    def __repr__(self):
        return f"ProofStep(conclusion=[{self.rule.consequent.label}] {' '.join(self.conclusion)})"

    def tree_string(self, label="", prefix=""):
        subst = {key: " ".join(val) for key, val in self.substitution.items()}
        ts = f"{prefix}[{label} <= {self.rule.consequent.label}] {' '.join(self.conclusion)} {subst}\n"
        for lab, dep in self.dependencies.items():
            ts += dep.tree_string(lab, prefix + " ")
        return ts

    # recursively collect all proof steps along the way to this one
    def all_steps(self, explored = None):

        # skip duplicated steps
        if explored == None: explored = set()
        if self.conclusion in explored: return []
        explored.add(self.conclusion)

        steps = [self]
        for dep in self.dependencies.values():
            steps.extend(dep.all_steps(explored))
        return steps

    # recursively build proof label sequence
    def normal_proof(self):

        # build proof if not cached
        if self._normal_proof is None:
            prf = ()
            for hyp in self.rule.hypotheses:
                prf += self.dependencies[hyp.label].normal_proof()
            prf += (self.rule.consequent.label,)
            self._normal_proof = prf

        # return proof
        return self._normal_proof

"""
Apply a rule to a sequence of dependencies, each a previous proof step 
The conclusions of each dependency should match the hypotheses of the rule
returns the resulting proof step and a status message
If the rule does not apply, proof step is None, otherwise status message is ""
"""
@profile
def perform(rule, dependencies):

    # make sure all hypotheses accounted for
    if len(rule.hypotheses) != len(dependencies):
        return None, f"{len(rule.hypotheses)} hypotheses != {len(dependencies)} dependences"

    # form substitution that unifies hypotheses with dependencies
    substitution = {}
    for (hypothesis, dependency) in zip(rule.hypotheses, dependencies):

        # update substitution for floating hypotheses
        if hypothesis.tag == "$f":

            # check matching types
            h_type, d_type = hypothesis.tokens[0], dependency.conclusion[0]
            if h_type != d_type:
                return None, f"{hypothesis.label}: mismatched types {h_type} vs {d_type}"

            # update substitution
            variable = hypothesis.tokens[1]
            substitution[variable] = dependency.conclusion[1:]

        # check that substitution unifies dependencies with essential hypotheses
        else: #if hypothesis.tag == "$e":
            substituted = substitute(hypothesis.tokens, substitution)
            if dependency.conclusion != substituted:
                substr = {k: " ".join(v) for k,v in substitution.items()}
                return None, f"{hypothesis.label}: {' '.join(dependency.conclusion)} != subst({' '.join(hypothesis.tokens)}, {substr})"

    # get all variables in substituted expressions and pairs of variables being substituted
    subvars = {v: rule.variables.intersection(tokens) for (v, tokens) in substitution.items()}
    variable_pairs = set(it.combinations(sorted(substitution.keys()), 2))

    # disjoint variable checks
    inherited = set()
    for (u, v) in variable_pairs & rule.disjoint:

        # check that disjoint variable substitutions remain disjoint
        if len(subvars[u] & subvars[v]) != 0:
            return None, f"{rule.consequent.label}: $d {u} {v} violated by {substitution}"

        # inherit disjoint requirements on the substituted variables
        for (x, y) in it.product(subvars[u], subvars[v]):
            inherited.add((min(x, y), max(x, y)))

    # infer conclusion from the rule
    conclusion = substitute(rule.consequent.tokens, substitution)

    # wrap dependencies in dictionary by hypothesis label
    dependencies = {hyp.label: dep for (hyp, dep) in zip(rule.hypotheses, dependencies)}

    # return results as a proof step with empty status message
    result = ProofStep(conclusion, rule, dependencies, substitution, inherited)
    return result, ""

"""
Conduct one step of a proof
Applies given rule to top of stack; pops top of stack in place
asserts disjoint variable requirements if provided
Returns the resulting proof step object and status message
If step is unsound, proof steps is None, otherwise message is ""
"""
@profile
def conduct(rule, stack, disjoint=None):

    # short-circuit hypothesis-less rules
    if len(rule.hypotheses) == 0:
        return ProofStep(tuple(rule.consequent.tokens), rule), ""

    # pop top of stack
    split = len(stack) - len(rule.hypotheses)
    dependencies = stack[split:]
    del stack[split:]

    # get next proof step by applying rule to top of stack
    # conclusion, substitution, disjoint = perform(rule, dependencies)
    step, msg = perform(rule, dependencies)
    if step is None: return step, msg

    # check against missing disjoint requirements
    if disjoint is not None:
        if step.disjoint > disjoint:
            return None, f"missing $d requirements: {step.disjoint} - {disjoint} = {step.disjoint - disjoint}"

    # return resulting proof step and normal status
    return step, ""

"""
claim: a rule object whose proof will be verified
returns root of proof tree and dictionary of proof step nodes
raises error if proof invalid
"""
@profile
def verify_normal_proof(database, claim):

    # initialize stack of proof steps
    stack = []

    # initialize set of proof steps, indexed by conclusion (tupled for hashability)
    proof_steps = {}

    # process each label in proof
    for step, step_label in enumerate(claim.consequent.proof):

        # print(f"\n{step}")
        # print(stack)

        # conduct next proof step
        rule = database.rules[step_label]
        proof_step, msg = conduct(rule, stack, claim.disjoint)
        assert msg == "", msg
        conclusion = proof_step.conclusion

        # save new proof steps
        if conclusion not in proof_steps:
            proof_steps[conclusion] = proof_step

        # push proof step onto stack
        stack.append(proof_step)

    # check that original claim has been proved
    assert len(stack) == 1, f"non-singleton stack {stack} after proof"
    assert stack[0].conclusion == tuple(claim.consequent.tokens), \
           f"proved statement {' '.join(stack[0].conclusion)} does not match theorem {' '.join(claim.consequent.tokens)}"

    # return root of proof graph and dictionary of nodes
    return stack[0], proof_steps

"""
claim: a rule object whose proof will be verified
returns root of proof tree and dictionary of proof step nodes
raises error if proof invalid
"""
@profile
def verify_compressed_proof(database, claim):

    # extract labels and mixed-radix pointer encodings
    split = claim.consequent.proof.index(")")
    step_labels = claim.consequent.proof[1:split]
    proof_string = ''.join(claim.consequent.proof[split+1:])

    # convert to integer pointers and save tagged steps
    A, U, Z = ord('A'), ord('U'), ord('Z')
    step_pointers = []
    pointer = 0
    for ordinal in map(ord, proof_string):
        if ordinal < U:
            pointer = 20 * pointer + (ordinal - A)
            step_pointers.append(pointer)
            pointer = 0
        elif ordinal < Z:
            pointer = 5 * pointer + (ordinal - U) + 1
        else:
            step_pointers.append(-1) # indicates previous step should be tagged

    # initialize proof stack
    stack = []

    # initialize buffer that gets dereferenced by step pointers
    proof_steps = []
    proof_step_dict = {}
    # steps for claim hypotheses
    for hypothesis in claim.hypotheses:
        conclusion, rule = tuple(hypothesis.tokens), database.rules[hypothesis.label]
        proof_steps.append(ProofStep(conclusion, rule))
        proof_step_dict[conclusion] = proof_steps[-1]
    # step labels
    proof_steps += step_labels

    # process each step in proof
    for step, pointer in enumerate(step_pointers):

        # tag previous step if requested
        if pointer < 0: proof_steps.append(stack[-1])

        # otherwise handle current proof step
        else:

            # retrieve current step
            proof_step = proof_steps[pointer]

            # replace labels by associated step
            if type(proof_step) is str:
                proof_step, msg = conduct(database.rules[proof_step], stack, claim.disjoint)
                assert msg == "", msg
                proof_step_dict[proof_step.conclusion] = proof_step

            # push current proof step onto stack
            stack.append(proof_step)

    # check that original claim has been proved
    assert stack[0].conclusion == tuple(claim.consequent.tokens), \
           f"proved statement {' '.join(stack[0].conclusion)} does not match theorem {' '.join(claim.consequent.tokens)}"
    assert len(stack) == 1, f"non-singleton stack {stack} after proof"

    # return root of proof graph and dictionary of nodes
    return stack[0], proof_step_dict

def verify_proof(database, claim):
    # compressed proofs start with "(" token
    if claim.consequent.proof[0] == "(":
        return verify_compressed_proof(database, claim)
    else:
        return verify_normal_proof(database, claim)

@profile
def verify_all(database, start=0, stop=-1):
    # verify all claims in the database
    for c, claim in enumerate(database.rules.values()):

        # only in specified slice
        if c < start: continue
        if c == stop: break

        # skip non-$p rules (axioms)
        if claim.consequent.tag != "$p": continue
        # print(c, claim.consequent.label)

        verify_proof(database, claim)

def normal_to_compressed():
    # or do this as post processing in mm?
    raise NotImplementedError

if __name__ == "__main__":

    from metamathpy.database import *
    # from .database import *

    # fpath = "p2.mm"
    # db = parse(fpath)

    # db.print()
    # claim = db.rules['mpd']
    # claim.print()

    # proof_root, proof_steps = verify_proof(db, claim)
    # proof_root.print(claim.consequent.label)

    # print("\nall nodes:\n")
    # keys = list(proof_steps.keys())
    # for k, conclusion in enumerate(keys):
    #     step = proof_steps[conclusion]
    #     edges = [keys.index(dep.conclusion) for dep in step.dependencies.values()]
    #     print(k, " ".join(conclusion), step.rule.consequent.label, edges)

    import os
    fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    # db = parse(fpath, last_rule = "bijust")
    db = parse(fpath)
    print("parsed.")

    # verify_all(db)
    # verify_all(db, stop=20000)
    # verify_compressed_proof(db, db.rules['ax5d'])

    # check proof string reconstruction
    for c, claim in enumerate(db.rules.values()):

        # skip non-$p rules (axioms)
        if claim.consequent.tag != "$p": continue

        # verify
        root, _ = verify_proof(db, claim)

        print(c, claim)
        print(claim.consequent.proof)

        # reconstruct normal proof and verify again
        stmt = claim.consequent
        proof = root.normal_proof()
        claim.consequent = Statement(stmt.label, stmt.tag, stmt.tokens, proof)
        print(claim.consequent.proof[:20])
        verify_normal_proof(db, claim)

    # # TODO: complete following demonstration of solitaire functionality

    # # conduct a proof, one step at a time, for x = x
    # # https://us.metamath.org/mmsolitaire/mms.html#q7
    # labels = ["ax-1", "ax-1", "ax-2", "ax-mp", "ax-mp", "ax-1", "ax-1", "ax-2", "ax-mp", "ax-mp",
    #     "impbi", # df-bi3 on metamath solitaire js
    #     "ax-mp", "ax-mp", "ax-gen", "ax-ext", "ax-mp",]

    # stack = []
    # for label in labels:
    #     rule = db.rules[label]

    #     # unify to get floating hypotheses (implicit in solitaire)
    #     unify(rule, stack)

    #     # for axioms, push their floating hypotheses onto the stack

    #     # conduct the step
    #     step = conduct(rule, stack)
    #     print(step)





from .setmm import load_pl
from .environment import Environment

def print_env(env):
    print("proof: " + " ".join(map(str, env.proof)))
    print("stack:")
    for s, step in enumerate(env.stack):
        print(" ", s, step)
    print(f"{len(env.proof)} steps so far...\n")

if __name__ == "__main__":

    db = load_pl()
    envs = [Environment(db)]
    
    while True:
        label = input("Enter label (q to quit, r to restart, u to undo, 0<i> for index): ")

        if label == "q":
            break

        if label == "r":
            envs = [Environment(db)]
            continue

        if label == "u":
            if len(envs) > 1:
                envs = envs[:-1]
                print_env(envs[-1])
            continue

        if label[0] != "0" and label not in db.rules:
            print("No rule called {label}, try again.")
            continue

        if label[0] == "0": label = int(label)

        env = envs[-1].copy()
        _, msg = env.step(label)

        if msg != "":
            print("Invalid rule: " + msg)
            print("try again.")
            continue

        envs.append(env)

        print_env(env)    



"""
Run from parent directory with
$ python -m tests.tests
"""
import os
import unittest as ut
from src.proof import substitute, verify_all
# from src.parse import parse as parse, Database
from src.database import parse as read

class TestSubstitute(ut.TestCase):

    def test_empty_sub(self):
        symbols = ("a", "b", "c")
        substitution = {}
        expected_result = ("a", "b", "c")
        actual_result = substitute(symbols, substitution)
        self.assertEqual(expected_result, actual_result)

    def test_complex_sub(self):
        symbols = ("a", "b", "c")
        substitution = {"a" : ('B -> C',), "c" : ('C -> D',)}
        expected_result = ("B -> C", "b", "C -> D")
        actual_result = substitute(symbols, substitution)
        self.assertEqual(expected_result, actual_result)

    def test_singleton_sub(self):
        symbols = ("a", "b", "c")
        substitution = {"a" : ('b',)}
        expected_result = ("b", "b", "c")
        actual_result = substitute(symbols, substitution)
        self.assertEqual(expected_result, actual_result)

# class TestParse(ut.TestCase):

#     def test_good_parse(self):
#         fpath = os.path.join('tests', 'p2.mm')
#         db = parse(fpath)
#         self.assertEqual(len(db.statements), 21)

#     def test_bad_parse(self):
#         fpath = os.path.join('tests', 'badparse.mm')
#         db = parse(fpath)
#         self.assertEqual(len(db.statements), 0)

#     def test_medium_parse(self):
#         fpath = os.path.join('tests', 'test.mm')
#         db = parse(fpath)
#         self.assertEqual(len(db.statements), 18)

#     def test_good_parse_rules(self):
#         fpath = os.path.join('tests', 'p2.mm')
#         db = parse(fpath)
#         self.assertEqual(len(db.rules), 10)

#     def test_bad_parse_rules(self):
#         fpath = os.path.join('tests', 'badparse.mm')
#         db = parse(fpath)
#         self.assertEqual(len(db.rules), 0)

#     def test_medium_parse_rules(self):
#         fpath = os.path.join('tests', 'test.mm')
#         db = parse(fpath)
#         self.assertEqual(len(db.rules), 9)

class TestVerify(ut.TestCase):
    def test_good_parse(self):
        fpath = os.path.join('tests', 'p2.mm')
        db = read(fpath)
        verify_all(db)

    def test_bad_parse(self):
        fpath = os.path.join('tests', 'badparse.mm')
        db = read(fpath)
        verify_all(db)

class TestDatabase(ut.TestCase):
    ## Bit confused on the Database version of parse, it always gives the same value for the rule parameter as statement however the Parse version gives seemingly a more accurate value
    def test_good_parse(self):
        fpath = os.path.join('tests', 'p2.mm')
        db = read(fpath)
        self.assertEqual(len(db.rules), 21)

    def test_bad_parse(self):
        fpath = os.path.join('tests', 'badparse.mm')
        db = read(fpath)
        self.assertEqual(len(db.rules), 0)

    def test_medium_parse(self):
        fpath = os.path.join('tests', 'test.mm')
        db = read(fpath)
        self.assertEqual(len(db.rules), 18)

if __name__ == '__main__':
    ut.main()



# run from mmpy with $ python -m src.mmmine.used
import sys
import os
import random
import pickle as pk
import itertools as it
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms

# proof length and size (uncompressed)
def proof_metrics(root):
    size = len(root.conclusion)
    leng = 1
    for dep in root.dependencies.values():
        dsize, dleng = proof_metrics(dep)
        size += dsize
        leng += dleng
    return size, leng

if __name__ == "__main__":

    do_count = False

    if do_count:

        # db = ms.load_pl()
        db = ms.load_all()

        theorem_labels = set()
        definition_labels = set()
        proof_lengths = {}
        form_lengths = {}
        label_counts = {}
        for r, (label, rule) in enumerate(db.rules.items()):
            print(f"rule {r} ({label}) of {len(db.rules)}...")

            # definitions are axioms
            if rule.consequent.tag == "$a" and label[:3] == "df-":
                definition_labels.add(label)

            # only look at proofs of propositions
            if rule.consequent.tag != "$p": continue

            # save labels of theorems
            theorem_labels.add(label)
    
            # uncompress proof
            root, steps = mp.verify_proof(db, rule)

            # track metrics
            # proof = root.normal_proof()
            # proof_lengths[len(proof)] = 1 + proof_lengths.get(len(proof), 0)
            proof_lengths[len(steps)] = 1 + proof_lengths.get(len(steps), 0)
            for step in steps.values():
                form_lengths[len(step.conclusion)] = 1 + form_lengths.get(len(step.conclusion), 0)

                lab = step.rule.consequent.label
                label_counts[lab]  = 1 + label_counts.get(lab, 0)

            # for lab in proof:
            #     label_counts[lab]  = 1 + label_counts.get(lab, 0)
    
        with open("used.pkl", "wb") as f:
            pk.dump((label_counts, theorem_labels, definition_labels, proof_lengths, form_lengths), f)

    with open("used.pkl", "rb") as f:
        label_counts, theorem_labels, definition_labels, proof_lengths, form_lengths = pk.load(f)

    import matplotlib.pyplot as pt
    from matplotlib import rcParams
    rcParams["font.family"] = "serif"

    # limit to labels of actual theorems and definitions
    thm_counts = {k:v for (k,v) in label_counts.items() if k in theorem_labels}
    def_counts = {k:v for (k,v) in label_counts.items() if k in definition_labels}

    print("theorems:")
    count_labels = sorted([(v,k) for (k,v) in thm_counts.items()])
    print(count_labels[:10])
    print(count_labels[-10:])

    print("definitions:")
    count_labels = sorted([(v,k) for (k,v) in def_counts.items()])
    print(count_labels[:10])
    print(count_labels[-10:])

    # print("form lengths:")
    # print(form_lengths)

    # calculate total formulation size
    formulation_size = 0
    for form_leng, count in form_lengths.items():
        formulation_size += form_leng * count
    print("Total formulation size:", formulation_size)

    # bucket the proof and form lengths
    pbs = 100
    for leng in sorted(proof_lengths.keys()):
        if leng % pbs == 0:
            continue
        num = proof_lengths.pop(leng)
        key = pbs * (leng // pbs)
        if key not in proof_lengths: proof_lengths[key] = 0
        proof_lengths[key] += num
    fbs = 350
    for leng in sorted(form_lengths.keys()):
        if leng % fbs == 0:
            continue
        num = form_lengths.pop(leng)
        key = fbs * (leng // fbs)
        if key not in form_lengths: form_lengths[key] = 0
        form_lengths[key] += num

    pt.figure(figsize=(8,2.5))
    pt.subplot(1,2,1)
    pt.plot(sorted(thm_counts.values()))
    pt.xlabel("Theorem")
    pt.ylabel("Usage")
    pt.yscale('log')
    pt.subplot(1,2,2)
    pt.plot(sorted(def_counts.values()))
    pt.xlabel("Definition")
    # pt.ylabel("Usage")
    pt.yscale('log')
    pt.tight_layout()
    pt.savefig("usage.eps")
    pt.show()

    pt.figure(figsize=(8,2.5))
    pt.subplot(1,2,1)
    pt.bar(list(proof_lengths.keys()), list(proof_lengths.values()), width=.8*pbs)
    pt.xlabel("Proof length")
    pt.ylabel("Frequency")
    pt.yscale('log')
    pt.subplot(1,2,2)
    pt.bar(list(form_lengths.keys()), list(form_lengths.values()), width=.8*fbs)
    pt.xlabel("Formula length")
    pt.yscale('log')
    # pt.xscale('log')
    pt.tight_layout()
    pt.savefig("sizing.eps")
    pt.show()




def hist(data):
    freqs = {}
    for val in data:
        freqs[val] = freqs.get(val, 0) + 1

    sorted_vals = sorted(freqs.keys())
    sorted_freqs = [freqs[val] for val in sorted_vals]
    print(f"{len(freqs)} unique vals, {np.min(sorted_vals)} to {np.max(sorted_vals)}")

    buckets = {}
    for val, freq in zip(sorted_vals, sorted_freqs):
        bucket = round(val, 1)
        buckets[bucket] = buckets.get(bucket, 0) + freq

    pt.plot(sorted_vals, sorted_freqs, 'k-')
    pt.plot(list(buckets.keys()), list(buckets.values()), 'bo-')

if __name__ == "__main__":

    import os
    import pickle as pk
    import numpy as np
    import matplotlib.pyplot as pt
    from database import parse
    from proof import verify_proof

    do_gen = True
    do_gen = False

    if do_gen:

        fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
        db = parse(fpath)
        print("parsed.")
    
        # lengths of rule consequents before and after substitutions
        sublens = []
        rule_labels = []
        for c, claim in enumerate(db.rules.values()):
    
            # if c >= 10000: break
    
            # skip non-$p rules (axioms)
            if claim.consequent.tag != "$p": continue
            print(c, claim.consequent.label)
    
            root, _ = verify_proof(db, claim)
            steps = root.all_steps()

            # grouped by claim
            sublens_c = []
            for step in steps:

                # grouped by step
                sublens_s = []

                sublens_s.append((len(step.rule.consequent.tokens), len(step.conclusion)))
                for lab, dep in step.dependencies.items():
                    sublens_s.append((len(db.statements[lab].tokens), len(dep.conclusion)))

                sublens_c.append(sublens_s)
            sublens.append(sublens_c)
            rule_labels.append(claim.consequent.label)

        with open("obvious.pkl", "wb") as f: pk.dump((sublens, rule_labels), f)

    with open("obvious.pkl", "rb") as f: (sublens, rule_labels) = pk.load(f)

    # ratios = [num/den for sublens_c in sublens for sublens_s in sublens_c for (num, den) in sublens_s]
    # hist(ratios)

    eqlens = [np.mean([
        np.min([before == after for (before, after) in sublens_s])
            for sublens_s in sublens_c])
                for sublens_c in sublens]
    idx = rule_labels.index('infpnlem2')
    print(f"infpnlem2: eqlen = {eqlens[idx]}")

    hist(eqlens)
    # pt.xscale('log')
    pt.yscale('log')
    pt.show()



# run from mmpy with $ python -m src.mmmine.scratch
import sys
import os
import random
import itertools as it
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import environment as me
from ..metamathpy import setmm as ms

if __name__ == "__main__":

    import matplotlib.pyplot as pt
    from matplotlib import rcParams
    rcParams["font.family"] = "serif"

    # db = ms.load_imp(fpath="/home/lightop/nexus/grants/nsf_aiming/set_id_normal.mm")
    db = ms.load_imp()
    # db = ms.load_ni()
    # db = ms.load_pl()
    # db.print()

    # proof length and size (uncompressed)
    def proof_metrics(root):
        size = len(root.conclusion)
        leng = 1
        for dep in root.dependencies.values():
            dsize, dleng = proof_metrics(dep)
            size += dsize
            leng += dleng
        return size, leng

    # get proof size for idALT
    # uncompressed: 40 steps, 232 size
    root, steps = mp.verify_proof(db, db.rules["idALT"])
    # size = sum(map(len, steps.keys())) # keys are conclusions
    size, leng = proof_metrics(root)
    for s, step in enumerate(steps.values()): print(s, " ".join(step.conclusion))
    input(f"idALT: {len(steps)} distinct step conclusions, {leng} proof length, {size} proof size...")

    # get axioms, split into wff and |-
    axioms = {"wff": {}, "|-": {}}
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$a":
            axioms[rule.consequent.tokens[0]][label] = rule

    # get all wff floating labels
    wff_labels = {}
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$f" and rule.consequent.tokens[0] == "wff":
            # maintain original sort order
            if label not in wff_labels: wff_labels[label] = 1
    print(wff_labels.keys())

    # get all wff variable labels appearing in proofs
    wff_vars = {}
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$p":
            root, _ = mp.verify_proof(db, rule)
            labels = set(wff_labels.keys()) & set(root.normal_proof())
            for label in labels:
                wff_vars[label] = db.rules[label]

    # sort wff vars for lexicographic enumeration
    # fl_labels = list(wff_vars.keys())
    # fl_labels = sorted(wff_vars.keys())
    wff_labels = tuple(label for label in wff_labels.keys() if label in wff_vars)
    wff_vars = {label: wff_vars[label] for label in wff_labels}

    print("axioms: " + " ".join(k for axs in axioms.values() for k in axs.keys()))
    print("metavars: " + " ".join(wff_labels))

    # make base proof steps for metavars
    wff_steps = {label: mp.perform(db.rules[label], ()) for label in wff_labels}
    print("wff steps: ", wff_steps)

    # # initialize base of theory dag, starting with floating wffs
    # wff_nodes = {}
    # for label, rule in wff_vars.items():
    #     step, msg = mp.perform(rule, ())
    #     proof = step.normal_proof()
    #     metavars = tuple(set(wff_labels) & set(proof))
    #     wff_nodes[proof] = (step, metavars)

    # print("\n*** wff nodes ***\n")
    # for proof, (step, metavars) in wff_nodes.items():
    #     print('step:', step)
    #     print('metavars:', metavars)
    #     print('proof:', proof)

    # input('.')

    # ent_nodes = {}

    # # apply one round of inference
    # new_nodes = {}

    # # try each rule
    # for rule in axioms["wff"].values():

    #     # try each combination of dependencies
    #     for fdeps in it.product(wff_nodes.values(), repeat=len(rule.floatings)):

    #         fdeps, metavars = zip(*fdeps)

    #         # skip combinations where metavars are not lexicographic
    #         lexo = True
    #         for d in range(1,len(metavars)+1):
    #             mv = set([v for mv in metavars[:d] for v in mv])
    #             if mv != set(wff_labels[:len(mv)]):
    #                 lexo = False
    #                 break
    #         if not lexo: continue

    #         # attempt rule on dependencies, continue if invalid
    #         node, msg = mp.perform(rule, fdeps)
    #         if msg != "": continue
            
    #         # save result
    #         new_nodes[node.normal_proof()] = node            

    #     # in raw graph, only ax-mp has essentials, so shortcut the possible dependencies

    # print("\n*** new nodes ***\n")
    # for proof, node in new_nodes.items():
    #     print('node:', node)
    #     print('proof:', proof)


    ### some stats about the database

    # for label, rule in db.rules.items():

    #     if rule.consequent.tag != "$p": continue
    #     num_props += 1

    #     for hyp in rule.essentials:
    #         toks = tuple(hyp.tokens)
    #         essentials[toks] = essentials.get(toks, 0) + 1
    #         essential_labels.append(hyp.label)

    #     print(label, rule)
    #     root, steps = mp.verify_proof(db, rule)
    #     for conclusion in steps.keys():
    #         wff = conclusion[1:]
    #         step_wffs[wff] = step_wffs.get(wff, 0) + 1

    #     proof_roots[label] = root
    #     proof_steps[label] = steps

    # print(f"\n*** wff vars ***\n")
    # for v, (label, rule) in enumerate(wff_vars.items()):
    #     print(v, rule)
    #     print(f" {len(rule.hypotheses)} hypotheses")

    # for key, axs in axioms.items():
    #     print(f"\n*** {key} ***\n")
    #     for a, (label, rule) in enumerate(axs.items()):
    #         print(a, rule)
    #         print(f" {len(rule.hypotheses)} hypotheses")

    # # wff BFS is just too large, even at very shallow depth (12 -> 6k -> oom)
    # # let's limit initial exploration as follows:
    # # only essential hypotheses appearing in set.mm pl are terminals for the raw |- graph
    # # only wffs appearing as proof steps in set.mm pl are nodes in the raw wff graph

    # essential_labels = []
    # essentials = {}
    # step_wffs = {}
    # proof_roots, proof_steps = {}, {}
    # num_props = 0
    # for label, rule in db.rules.items():

    #     if rule.consequent.tag != "$p": continue
    #     num_props += 1

    #     for hyp in rule.essentials:
    #         toks = tuple(hyp.tokens)
    #         essentials[toks] = essentials.get(toks, 0) + 1
    #         essential_labels.append(hyp.label)

    #     print(label, rule)
    #     root, steps = mp.verify_proof(db, rule)
    #     for conclusion in steps.keys():
    #         wff = conclusion[1:]
    #         step_wffs[wff] = step_wffs.get(wff, 0) + 1

    #     proof_roots[label] = root
    #     proof_steps[label] = steps

    # print(f"{len(essentials)} distinct essential hypotheses across {num_props} props")
    # print(f"{len(step_wffs)} distinct wffs across {num_props} proofs")

    # print("least|most frequent hypotheses:")
    # sort_ess = sorted([(v,k) for (k,v) in essentials.items()])
    # print(" ".join(sort_ess[0][1]))
    # print(" ".join(sort_ess[-1][1]))

    # print("least|most frequent wffs:")
    # sort_wffs = sorted([(v,k) for (k,v) in step_wffs.items()])
    # print(" ".join(sort_wffs[0][1]))
    # print(" ".join(sort_wffs[-1][1]))

    # input('..')

    # fig, axs = pt.subplots(1,2, figsize=(6.5,2.5))
    # # axs[0].bar(range(len(essentials)), sorted(essentials.values()))
    # axs[0].plot(sorted(essentials.values()))
    # axs[0].set_xlabel("Essential hypothesis")
    # axs[0].set_ylabel("Occurrences in theorems")
    # axs[0].set_yscale('log')
    # # axs[1].bar(range(len(step_wffs)), sorted(step_wffs.values()))
    # axs[1].plot(sorted(step_wffs.values()))
    # axs[1].set_xlabel("Well-formed formula")
    # axs[1].set_ylabel("Occurences in proofs")
    # axs[1].set_yscale('log')
    # pt.tight_layout()
    # # pt.savefig("wff_freqs.png", dpi=600)
    # pt.savefig("wff_freqs.eps")
    # pt.show()

    # fig, axs = pt.subplots(1,1, figsize=(6.5,2.5))
    # axs = [axs] # 1,1
    # axs[0].hist(list(map(len, proof_steps.values())))
    # axs[0].set_xlabel("Proof length")
    # axs[0].set_ylabel("Frequence")
    # pt.tight_layout()
    # pt.savefig("proof_lens.eps")
    # pt.show()

    ### still having trouble figuring out how to organize the search effectively.
    ### different approach: forward search on valid proof strings.

    ax_labels = tuple(list(axioms['wff'].keys()) + list(axioms["|-"].keys()))

    env = me.Environment(db)

    envs = {0: [env]}

    max_depth = int(sys.argv[1]) #11
    for depth in range(max_depth):
        envs[depth+1] = []

        # try applying rules to each environment
        for e,env in enumerate(envs[depth]):

            # try every axiom
            try_labels = ax_labels

            # can also push previous proof steps
            try_labels += tuple(n for (n,lab) in enumerate(env.proof) if type(lab) is str)

            # can also push a new metavariable floating hypothesis, in lexicographic order
            fl_idx = len(set(env.proof) & set(wff_labels))
            try_labels += wff_labels[fl_idx:fl_idx+1]

            # try applying each rule
            for l,label in enumerate(try_labels):

                # skip envs with no hope of completing by max depth.
                # ax-mp is most-reducing rule; pops 4 and pushes 1, net -3
                # can only complete if current length <= 1 + 3*(max_depth-depth)
                # at max_depth-depth==1, at most 4 in stack
                # at max_depth-depth==2, at most 7 in stack
                # etc
                if len(env.stack) > 1 + 3*(max_depth - depth): continue

                # copy env and step
                new_env = env.copy()
                _, msg = new_env.step(label)

                # skip invalid steps
                if msg != "": continue

                # skip non-int labels that reconstruct a previous proof step
                if type(label) is str:
                    step_proofs = [step.normal_proof() for step in new_env.steps]
                    if step_proofs[-1] in set(step_proofs[:-1]): continue

                # save new environment
                envs[depth+1].append(new_env)
                # print(f" env, label {e}, {l} of {len(envs[depth]), len(ax_labels)}")
            # print(f" env {e} of {len(envs[depth])}")

        print(f"depth {depth+1} envs: {len(envs[depth+1])}")

    for depth, envs_d in envs.items():
        print(f"depth {depth}: {len(envs_d)} envs")


    # ## OLD approach without proof backpointers    

    # # ax_labels = essential_labels + list(axioms['wff'].keys()) + list(axioms["|-"].keys())
    # ax_labels = list(axioms['wff'].keys()) + list(axioms["|-"].keys())
    # wff_labels = list(wff_vars.keys()) # uncompressed proofs also push $f statements

    # env = me.Environment(db)

    # envs = {0: [env]}

    # max_depth = 6
    # for depth in range(max_depth):
    #     envs[depth+1] = []

    #     # try applying rules to each environment
    #     for e,env in enumerate(envs[depth]):

    #         # set up labels to try, don't skip floatings
    #         fl_so_far = set(env.proof) & set(wff_labels)
    #         if len(fl_so_far) == 0:
    #             fl_idx = 1
    #         else:
    #             fl_idx = max(map(wff_labels.index, fl_so_far))+2
    #         try_labels = ax_labels + wff_labels[:fl_idx]

    #         # try applying each rule
    #         for l,label in enumerate(try_labels):

    #             # skip envs with no hope of completing by max depth.
    #             # ax-mp is most-reducing rule; pops 4 and pushes 1, net -3
    #             # can only complete if current length <= 1 + 3*(max_depth-depth)
    #             # at max_depth-depth==1, at most 4 in stack
    #             # at max_depth-depth==2, at most 7 in stack
    #             # etc
    #             if len(env.stack) > 1 + 3*(max_depth - depth): continue

    #             # copy env and step
    #             new_env = env.copy()
    #             _, msg = new_env.step(label)

    #             # skip invalid steps
    #             if msg != "": continue

    #             # save new environment
    #             envs[depth+1].append(new_env)
    #             # print(f" env, label {e}, {l} of {len(envs[depth]), len(ax_labels)}")
    #         # print(f" env {e} of {len(envs[depth])}")

    #     print(f"depth {depth+1} envs: {len(envs[depth+1])}")

    # for depth, envs_d in envs.items():
    #     print(f"depth {depth}: {len(envs_d)} envs")

    # # Depth & Valid partial proofs \\
    # # 1 & 6 \\
    # # 2 & 42 \\
    # # 3 & 402 \\
    # # 4 & 3822 \\
    # # 5 & 35358 \\
    # # 6 & 326562 \\
    # # 7 & 3023610 \\
    # # 8 & 13806006

    # # 0 & 1 \\
    # # 1 & 1 \\
    # # 2 & 3 \\
    # # 3 & 16 \\
    # # 4 & 86 \\
    # # 5 & 485 \\
    # # 6 & 2935 \\
    # # 7 & 18956 \\
    # # 8 & 129553 \\
    # # 9 & 931517 \\
    # # 10 & 6467657 \\
    # # 11 & 14119609

    # # 0 & 1 \\
    # # 1 & 1 \\
    # # 2 & 2 \\
    # # 3 & 9 \\
    # # 4 & 40 \\
    # # 5 & 192 \\
    # # 6 & 1000 \\
    # # 7 & 5636 \\
    # # 8 & 33929 \\
    # # 9 & 216220 \\
    # # 10 & 1447866 \\
    # # 11 & 6576611 \\
    # # 12 & 7564433 


    # extract all complete proofs
    proved = {}
    proofs = {}
    psizes = {}
    plengs = {}
    for depth, envs_d in envs.items():
        for e, env in enumerate(envs_d):

            # complete means exactly one left on stack
            if len(env.stack) != 1: continue

            (step,) = env.stack
            conc = step.conclusion
            if conc not in proved: proved[conc] = []
            proved[conc].append((depth,e))
            proofs[conc] = step.normal_proof()
            psizes[conc], plengs[conc] = proof_metrics(step)

    ents = {k: v for (k,v) in proved.items() if k[0] == "|-"}

    print(f"{len(proved)} distinct conclusions proved")
    print(f"{len(ents)} are |- statements")
    print("some random ones:")
    for p, (conc, pts) in enumerate(random.sample(list(ents.items()), min(len(ents), 10))):
        deps, idxs = zip(*pts)
        print(p, " ".join(conc), list(deps), idxs)

    print("how many with exactly 1 proof?")
    print(len([v for v in ents.values() if len(v) == 1]))
    print("how many with > 1 proof?")
    print(len([v for v in ents.values() if len(v) > 1]))
    print("how many with different length proofs?")
    print(len([v for v in ents.values() if len(set(list(zip(*v))[0])) > 1]))
    print(f"largest size = {max(psizes.values())}, leng = {max(plengs.values())}")

    print("how many complete proofs at each depth?")
    proofs_at = {}
    for (conc, pts) in ents.items():
        deps, idxs = zip(*pts)
        for dep in deps:
            proofs_at[dep] = proofs_at.get(dep, 0) + 1

    for dep in sorted(proofs_at.keys()): print(dep, proofs_at[dep])

    # 3 & 2 \\
    # 4 & 5 \\
    # 5 & 12 \\
    # 6 & 62 \\
    # 7 & 119 \\
    # 8 & 936 \\
    # 9 & 1659 \\
    # 10 & 17612 


    # show two proofs for same claim
    print("2 different proofs:")
    for (conc, pts)in ents.items():
        if len(pts) == 1: continue

        print(" ".join(conc))
        for d,e in pts:
            print(envs[d][e].proof)        

        break
    print()

    # check that as long as there are no claims with multiple proofs, is every proof_step a distinct conclusion?
    all_steps = {} # memory address -> step object
    for depth, envs_d in envs.items():
        for e, env in enumerate(envs_d):
            for step in env.steps:
                all_steps[id(step)] = step
                # print(depth,e,id(step),step)

    all_concs = set()
    for sid, step in all_steps.items():
        all_concs.add(step.conclusion)
        # print(sid, " ".join(step.conclusion))

    # This check is wrong!! different environments can duplicate "wff ps" used at different times during the different proofs.
    print(f"Surprise: {len(all_concs)} distinct conclusions != {len(all_steps)} distinct steps in memory")

    # print("the normal proofs themselves:")
    # for (conc, proof) in proofs.items():
    #     print(" ".join(conc))
    #     print("   ", " ".join(proof))

    # for each proof, count how many other proofs use it.
    # another proof "uses" it if its normal proof is a substring of the other proof,
    # and the other proof has one more stack entry after the subsequence than before
    # in fact, given that it itself is a valid proof, the second condition is already implied.
    # (of course, shallower steps will be used by more, but within layer?)
    # (can you say anything stronger past max_depth?  will all proof steps eventually be used equally?)
    # (something like the exponential(?) growth rate of superproofs that use each theorem?)

    print("\n*** depth containment relations:\n")

    subdeps, superdeps = {}, {}
    for (conc, prf1), prf2 in it.product(proofs.items(), proofs.values()):
        if len(prf1) >= len(prf2): continue
        
        if conc not in subdeps:
            subdeps[conc] = len(prf1)
            superdeps[conc] = []

        if " ".join(prf1) in " ".join(prf2):
            superdeps[conc].append(len(prf2))

    # for conc in subdeps:
    #     # if len(superdeps[conc]) == 0: continue
    #     # if conc[0] == "wff": continue
    #     print(subdeps[conc], ": ", " ".join(conc))
    #     print("is in:", superdeps[conc])

    # scatdat = np.array([(subdeps[conc], d) for conc in subdeps for d in superdeps[conc]])
    # x, y = (scatdat + .5*np.random.rand(*scatdat.shape)).T
    # pt.plot(x, y, 'k.')

    histdat = np.array([(subdeps[conc], len(superdeps[conc])) for conc in subdeps])
    deps = [d for d in range(2,max_depth+1) if (histdat[:,0] == d).any() and (histdat[histdat[:,0]==d,1] > 0).any()]
    fig, axs = pt.subplots(1, len(deps), figsize=(6.5,2))
    for d,dep in enumerate(deps):
        idx = histdat[:,0] == dep
        axs[d].hist(histdat[idx,1])#, bins = np.arange(histdat[idx,1].max()+1)-.5)
        axs[d].set_title(f"Proof length {dep}")
        if d > 0: axs[d].set_yscale("log")
    fig.supxlabel("Number of superproofs")
    fig.supylabel("Frequency")
    pt.tight_layout()
    pt.savefig("superproofs.eps")
    pt.show()

    # containers = {}
    # for d1 in range(max_depth);
    #     for env1 in envs[d1]:
    #         (step,) = env1.stack
    #         subconc = step.conclusion
    #         subproof = step.normal_proof()
    #         for d2 in range(1,max_depth+1):
    #             for env2 in envs[d2]:
    #                 (step,) = env2.stack
    #                 superproof = step.normal_proof()

    # # BFS on wffs
    # max_depth = 2
    # explored_wffs = {}
    # for label, rule in wff_vars.items():
    #     tokens = tuple(rule.consequent.tokens)
    #     explored_wffs[" ".join(tokens)] = mp.ProofStep(tokens, rule)

    # print(f"\n *** depth 0: {len(explored_wffs)} ***")
    # print(explored_wffs)

    # for depth in range(max_depth):
    #     more_wffs = {}
    #     for (label, rule) in axioms["wff"].items():
    #         hyps = len(rule.hypotheses)
    #         for deps in it.product(explored_wffs.values(), repeat=hyps):
    #             step, msg = mp.perform(rule, deps)
    #             assert msg == ""
    #             more_wffs[" ".join(step.conclusion)] = step
    #     explored_wffs.update(more_wffs)

    #     print(f"\n *** depth {depth+1}: {len(explored_wffs)} ***")
    #     # print(explored_wffs)

    # # a = 0
    # # for r, (label, rule) in enumerate(db.rules.items()):
    # #     if rule.consequent.tag != "$a": continue
    # #     print(a, rule)
    # #     a += 1

    # # db.dump("")
    
    # # proof_roots = {}
    # # proof_steps = {}
    
    # # # According to set.mm, prop logic has 194 axioms, definitions, and theorems, but this looks to be outdated.
    # # # last label before any FOL (universal quantifier) is xorexmid, it is "rule" 2849 (including hypotheses)
    # # last_pl = "xorexmid"
    
    # # for r, (label, rule) in enumerate(db.rules.items()):
    
    # #     if rule.consequent.tag == "$p":
    # #         print(r,label,rule)
    # #         root, steps = mp.verify_proof(db, rule)
    # #         proof_roots[label] = root
    # #         proof_steps[label] = steps
    
    # #     if label == last_pl: break
    
    
    # # print('num pl', num_pl)
    # # # stats on number of proof steps?
    # # print('num proofs', len(proof_steps))
    # # print('shortest proof', min(map(len, proof_steps.values())))
    # # print('longest proof', max(map(len, proof_steps.values())))
    # # print('total proof steps', sum(map(len, proof_steps.values())))
    
    # # # how many proof steps are redundant?
    # # distinct_steps = {}
    # # for label, steps in proof_steps.items():
    # #     hyps = tuple(" ".join(hyp.tokens) for hyp in db.rules[label].hypotheses)
    # #     for step_conclusion in steps.keys():
    # #         key = (" ".join(step_conclusion),) + hyps
    # #         if key not in distinct_steps: distinct_steps[key] = []
    # #         distinct_steps[key].append(label)
    
    # # print('distinct', len(distinct_steps))
    
    



from line_profiler import profile
import numpy as np
import pickle as pk
from time import perf_counter
import itertools as it
import metamathpy.proof as mp
from metamathpy.unification import substitute

class Pile:

    # TODO?
    # - cache all valid rule applications to avoid re-checking?
    # - make rules, steps immutable and store in sets
    # - save all proof steps with same conclusion
    # - avoid repeating a previous rule application within dfs path (new rules for each proof step?)
    # - avoid duplicating same final proof steps in different orders across dfs paths (lexicographic floating introduction?)
    # - make sure you avoid any circular proofs (eg shorter path to id using things proved after it)

    def __init__(self, db=None, ax_only=False):
        self.rules = []
        # proof steps by type, keyed by conclusion
        self.steps = {"wff": {}, "|-": {}}
        self.trace = []

        if db is None: return

        # filter rules
        for rule in db.rules.values():

            # for now, omit "rules" for essential hypotheses
            if rule.consequent.tag == "$e": continue

            # if requested, omit any non-axiom propositions
            if ax_only and rule.consequent.tag == "$p": continue

            # otherwise keep the rule
            self.rules.append(rule)

        # maintain a record of valid rule applications?

    def __str__(self):
        s = "Pile:\n Rules:\n"
        for rule in self.rules:
            s += f" {' '.join(rule.consequent.tokens)}\n"
        s += "Steps:\n"
        # for step in list(self.steps["wff"].values()) + list(self.steps["|-"].values()):
        #     s += f" {step}\n"
        for step in self.trace: s += f" {step}\n"
        return s

    def copy(self):
        other = Pile()
        other.rules = list(self.rules)
        other.steps = {k: dict(v) for (k,v) in self.steps.items()}
        other.trace = list(self.trace)
        return other

    def contained_in(self, background):
        return all([
            step.conclusion in background
            for step in self.trace])

    @profile
    def children(self):
        """
        Use with care; yields in-place modifications of self as each child
        """
        # try applying each rule
        for rule in self.rules:

            if len(rule.hypotheses) == 0:

                step, status = mp.perform(rule, ())
                assert status == "" # no hypotheses, should always be valid

                tokens = step.conclusion
                typecode = tokens[0] # wff or |-
                if tokens not in self.steps[typecode]:

                    self.steps[typecode][tokens] = step
                    self.trace.append(step)

                    yield self
    
                    # *** undo changes for next child, careful with this
                    self.steps[typecode].pop(tokens)
                    self.trace.pop()

            else:

                # apply rule to all wff combinations for metavariables
                wffs = list(self.steps["wff"].values())
                for deps in it.product(wffs, repeat=len(rule.floatings)):
                    print(deps)

                    # set up substitution
                    substitution = {}
                    for (hyp, dep) in zip(rule.floatings, deps):
                        # assert matching types (only works for PL)
                        assert hyp.tokens[0] == dep.conclusion[0] == "wff"
                
                        # update substitution
                        variable = hyp.tokens[1]
                        substitution[variable] = dep.conclusion[1:]

                    # apply substitution and check if essentials are satisfied
                    satisfied = True
                    for hyp in rule.essentials:
                        substituted = substitute(hyp.tokens, substitution)
                        if substituted not in self.steps["|-"]:
                            satisfied = False
                            break
                        deps = deps + (self.steps["|-"][substituted],)

                    # if not, move on to next combo
                    if not satisfied: continue

                    print(deps)
                    input('.')

                    # otherwise, get conclusion and make sure inference checks out
                    step, status = mp.perform(rule, deps)
                    assert status == ""

                    # append resulting proof step while yielding this child
                    tokens = step.conclusion
                    typecode = tokens[0] # wff or |-
                    if tokens not in self.steps[typecode]:
                        self.steps[typecode][tokens] = step
                        self.trace.append(step)
    
                        yield self
        
                        # *** undo changes for next child, careful with this
                        self.steps[typecode].pop(tokens)
                        self.trace.pop()

                # # rule signature: floating and essential
                # wffs = list(self.steps["wff"].values())
                # ents = list(self.steps["|-"].values())
                # feeds = (wffs,)*len(rule.floatings) + (ents,)*len(rule.essentials)
                # for deps in it.product(*feeds):

                #     step, status = mp.perform(rule, deps)

                #     # skip invalid rule applications
                #     if status != "": continue

                #     tokens = step.conclusion
                #     typecode = tokens[0] # wff or |-
                #     if tokens not in self.steps[typecode]:
                #         self.steps[typecode][tokens] = step
                #         self.trace.append(step)
    
                #         yield self
        
                #         # *** undo changes for next child, careful with this
                #         self.steps[typecode].pop(tokens)
                #         self.trace.pop()

        return

def dl_dfs(pile, max_depth=0):

    # check if max depth reached
    if max_depth <= 0:
        yield pile
        return

    # # skip if contained in background
    # if pile.contained_in(background):
    #     return

    # add to background

    # recurse search from each child
    for child in pile.children():
        yield from dl_dfs(child, max_depth=max_depth-1)

if __name__ == "__main__":

    import metamathpy.setmm as ms

    loader = ms.load_ni
    # loader = ms.load_pl
    # loader = lambda : ms.load_to("mpd")
    ax_only = False
    max_depth = 1
    max_time = np.inf#1 * 60 * 60

    # question: for a given max depth, how many set.mm theorems can you validly re-prove (without using their dependents)
    # quick way before piles add rules for proof steps: recursively find rules used in the proof step's proof and make sure all come before the re-proved one

    db = loader()
    print(db)

    pile = Pile(db, ax_only)

    # # quick id check: remove all unneeded rules
    # pile.rules = [r for r in pile.rules if r.consequent.label in ("wph", "wi", "ax-1", "mpd")]

    # print(pile)
    # input('..')

    # setup goals to prove
    goals = {}
    rule_labels = list(db.rules.keys())
    for n, (label, rule) in enumerate(db.rules.items()):
        # only look at propositions
        if rule.consequent.tag != "$p": continue

        # and only with no essential hypotheses
        if len(rule.essentials) > 0: continue

        # save conclusion and position in db
        goals[tuple(rule.consequent.tokens)] = (label, n)

    # get all goal conclusion sizes
    goal_sizes = set()
    for tok in goals.keys():
        goal_sizes.add(len(tok)-1) # -1 omits typecode

    # print("goal sizes:")
    # print(sorted(goal_sizes))
    # input('.')

    reproved = {}
    all_steps = {"wff": {}, "|-": {}}
    conc_freqs = {}
    num_leaves = 0
    start = perf_counter()
    for leaf in dl_dfs(pile, max_depth=max_depth):
        num_leaves += 1

        # check if a goal was proved
        for toks, step in leaf.steps["|-"].items():
            if toks in goals:
                _, n = goals[toks]
                proof = step.normal_proof()
                # reproved[toks] = proof
                if set(proof) <= set(rule_labels[:n]):
                    reproved[toks] = proof

        # # track proof step frequencies
        # for conc in list(leaf.steps["wff"].values()) + list(leaf.steps["|-"].values()):
        #     conc_freqs[conc] = conc_freqs.get(conc, 0) + 1

        # print(leaf)
        # print(conc_freqs)
        # input('.')

        # # get coefficients
        # typed_rules = {"wff":[], "|-":[]}
        # all_coefs = {"wff":[], "|-":[]}
        # for rule in leaf.rules:
        #     if rule.consequent.tag == "$f": continue

        #     typecode = rule.consequent.tokens[0] # organize by consequent typecode

        #     wffs = [f.tokens[1] for f in rule.floatings] # omit floating typecodes
        #     coefs = np.array([rule.consequent.tokens.count(wff) for wff in wffs])
        #     const = sum([t not in wffs for t in rule.consequent.tokens[1:]]) # omit typecode

        #     typed_rules[typecode].append(rule)
        #     all_coefs[typecode].append( (coefs, const) )

        #     # print(rule)
        #     # print(" + ".join([f"{c}|{w}|" for (c,w) in zip(coefs,wffs)] + [str(const)]))
        #     # input('.')

        # # get currently built wff sizes
        # wff_sizes = set()
        # for toks, step in leaf.steps["wff"].items():
        #     wff_sizes.add(len(toks)-1) # -1 omits typecode
        # print("wff sizes:")
        # print(sorted(wff_sizes))

        # # get n-hop sizes
        # hop_sizes = {"wff": set(wff_sizes), "|-": set()}
        # for hop in range(1):
        #     combos = {}
        #     next_hop_sizes = {}
        #     for typecode in ("wff", "|-"):
        #         next_hop_sizes[typecode] = set()

        #         for (rule, (coefs, const)) in zip(typed_rules[typecode], all_coefs[typecode]):
        #             if len(coefs) not in combos:
        #                 combos[len(coefs)] = np.array(tuple(it.product(hop_sizes["wff"], repeat=len(coefs))))

        #             print(rule)
        #             print(" + ".join([f"{c}|{w}|" for (c,w) in zip(coefs,wffs)] + [str(const)]))
        #             print("combos:")
        #             print(combos[len(coefs)])
        #             print("coefs:")
        #             print(coefs, const)
        #             print("combos @ coefs:")
        #             print(set((combos[len(coefs)] @ coefs + const).astype(int)))
        #             input('.')

        #             next_hop_sizes[typecode] |= set((combos[len(coefs)] @ coefs + const).astype(int))

        #     print(f"hop {hop+1} sizes:")
        #     for typecode in ("wff", "|-"):
        #         hop_sizes[typecode] |= next_hop_sizes[typecode]
        #         print(typecode, sorted(hop_sizes[typecode]))
        #     print("vs goal sizes:")
        #     print(sorted(goal_sizes))

        #     input('.')

        duration = perf_counter() - start
        if duration > max_time: break

    # # print(leaf)
    # unifreqs = set(conc_freqs.values())
    # freq_freqs = {f: list(conc_freqs.values()).count(f) for f in unifreqs}
    # print(freq_freqs)
    # for conc, freq in conc_freqs.items():
    #     if freq == max(unifreqs): print(conc)
    # # for conc, freq in conc_freqs.items():
    # #     print(freq, conc)

    # print(f"\nDistinct wffs:")
    # for c, conc in enumerate(all_steps["wff"].keys()): print(c, " ".join(conc))
    # print(f"\nDistinct |-:")
    # for c, conc in enumerate(all_steps["|-"].keys()): print(c, " ".join(conc), conc)

    # si_tok = ("|-", "(", "si", "->", "(", "-.", "si", "->", "ta", ")", ")")
    # et_tok = ("|-", "(", "(", "(", "et", "->", "et", ")", "->", "si", ")", "->", "(", "-.", "et", "->", "si", ")", ")")
    # id_tok = ("|-", "(", "ph", "->", "ph", ")")

    # # print("\nlast leaf:\n")
    # # print(last)
    # print(f"\n{num_leaves} leaves total at depth {max_depth}, {len(all_steps['wff'])} wffs and {len(all_steps['|-'])} |-s")
    # print(f"si? {si_tok in all_steps['|-']}")
    # print(f"et? {et_tok in all_steps['|-']}")
    # print(f"id? {id_tok in all_steps['|-']}")

    for n, (toks, proof) in enumerate(reproved.items()):
        lab, _ = goals[toks]
        step, _ = mp.verify_compressed_proof(db, db.rules[lab])
        print(f"{n: 3d}  |np|={len(proof)} [{lab}] {' '.join(toks)} <{' '.join(proof)}>  VS  <{' '.join(step.normal_proof())}>")
    print(f"{len(reproved)} of {len(goals)} reproved")

    with open(f"ida_d{max_depth}.pkl", "wb") as f: pk.dump(reproved, f)

    print(f"Took {duration:.3f}s ({duration/60:.3f}min, {duration/(60*60):.3f}hr)")




import os
from database import *
# from parse import *

def substitute(symbols, substitution):
    result = []
    for symbol in symbols:
        result.extend(substitution.get(symbol, [symbol]))
    return result

# one inference step in a proof
class Inference:
    def __init__(self, db, rule, dependencies, substitution):
        self.rule = rule # justification for the step
        self.dependencies = dependencies # previous steps it relies on
        self.substitution = substitution # substitution that matches them

        # form conclusion from substitution
        consequent = db.statements[rule.consequent]
        self.conclusion = substitute(consequent.tokens, substitution)

    def __repr__(self):
        return f"{self.rule.consequent} {' '.join(self.conclusion)}"
    def print(self, label="", prefix=""):
        subst = {key: " ".join(val) for key, val in substitution.items()}
        # print(f"{prefix}[{label} <= {self.rule.consequent}] {' '.join(self.conclusion)} <= {' '.join(self.rule.symbols)}")
        print(f"{prefix}[{label} <= {self.rule.consequent}] {' '.join(self.conclusion)} {subst}")
        for lab, dep in self.dependencies.items():
            dep.print(lab, prefix + " ")

if __name__ == "__main__":

    # fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    fpath = "p2.mm"
    db = parse(fpath)
    db.print()
    # thm = db.get_statement('syl')
    thm = db.rules['mpd']
    stmt = db.statements[thm.consequent]
    thm.print()

    print("\n" + "*"*8 + "\n")

    # build uncompressed proof
    stack = []
    for s,label in enumerate(stmt.proof):
        statement = db.statements[label]
        print(f"\n{s}")
        print(stack)

        if statement.tag in ("$f", "$e"):
            rule = Rule(db, label, [], [])
            inf = Inference(db, rule, {}, {})
            stack.append(inf)

        if statement.tag in ("$a", "$p"):

            # look up rule whose consequent is statement
            rule = db.rules[label]

            # pop dependencies from stack
            hypotheses = rule.floatings + rule.essentials
            split = len(stack) - len(hypotheses)
            stack, dependencies = stack[:split], stack[split:]
            substitution = {}

            for dep, hyp in zip(dependencies, hypotheses):
                hypothesis = db.statements[hyp]
                dependency = dep.conclusion

                if hypothesis.tag == "$f":
                    r_type, r_sym = hypothesis.tokens[0], hypothesis.tokens[1:]
                    s_type, s_sym = dependency[0], dependency[1:]
                    assert len(r_sym) == 1, \
                           f"step {s} {label}: non-variable floating dependency {r_sym}"
                    assert r_type == s_type, \
                           f"step {s} {label}: mismatched types {r_type} vs {s_type}"
                    substitution[r_sym[0]] = s_sym

                if hypothesis.tag == "$e":
                    assert dependency == substitute(hypothesis.tokens, substitution), \
                           f"step {s} {label}: {dependency} != subst({hypothesis.tokens}, {substitution})"

            dependencies = {hyp: dep for (hyp, dep) in zip(hypotheses, dependencies)}
            inf = Inference(db, rule, dependencies, substitution)
            stack.append(inf)

    print(f"\nend:")
    print(stack)

    assert len(stack) == 1, f"non-singleton stack {stack} after proof"
    assert stack[0].conclusion == stmt.tokens, \
           f"proved statement {stack[0].conclusion} does not match theorem {stmt.tokens}"

    root_inf = stack[0]
    root_inf.print(stmt.label)


    # # fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    # fpath = "p2.mm"
    # db = parse(fpath)
    # db.print()
    # # thm = db.get_statement('syl')
    # thm = db.get_statement('mpd')
    # print(thm.label, thm.tag, thm.symbols)

    # # get scope
    # scope = thm.block.get_scope(thm)
    # print(scope)

    # print("\n" + "*"*8 + "\n")

    # # build uncompressed proof
    # stack = []
    # for s,label in enumerate(thm.proof):
    #     rule = db.get_rule(label)
    #     conclusion = db.get_statement(label)
    #         print(f"\n{s}")
    #     print(stack)
    #     print(conclusion)

    #     if type(conclusion) == Hypothesis:
    #         inf = Inference(conclusion, {}, {})
    #         stack.append(inf)

    #     if type(conclusion) in (Axiom, Proposition):

    #         scope = conclusion.block.get_scope(conclusion)

    #         # remove floating hypotheses for irrelevant variables
    #         variables = set(conclusion.symbols)
    #         for e in scope.e: variables.update(e.symbols)
    #         variables.intersection_update(scope.v)
    #         floats = [f for f in scope.f if f.symbols[1] in variables]

    #         # pop dependencies from stack
    #         hypotheses = floats + scope.e
    #         split = len(stack) - len(hypotheses)
    #         stack, dependencies = stack[:split], stack[split:]
    #         substitution = {}

    #         for dep, hyp in zip(dependencies, hypotheses):

    #             if hyp.tag == "$f":
    #                 r_type, r_sym = hyp.symbols[0], hyp.symbols[1:]
    #                 s_type, s_sym = dep.conclusion[0], dep.conclusion[1:]
    #                 assert len(r_sym) == 1, \
    #                        f"step {s} {label}: non-variable floating dependency {r_sym}"
    #                 assert r_type == s_type, \
    #                        f"step {s} {label}: mismatched types {r_type} vs {s_type}"
    #                 substitution[r_sym[0]] = s_sym

    #             if hyp.tag == "$e":
    #                 assert dep.conclusion == substitute(hyp.symbols, substitution), \
    #                        f"step {s} {label}: {s_hyp.symbols} != subst({hyp.symbols}, {substitution})"

    #         dependencies = {hyp.label: dep for (hyp, dep) in zip(hypotheses, dependencies)}
    #         inf = Inference(conclusion, dependencies, substitution)
    #         stack.append(inf)

    # print(f"\nend:")
    # print(stack)

    # assert len(stack) == 1, f"non-singleton stack {stack} after proof"
    # assert stack[0].conclusion == thm.symbols, f"proved statement {stack[0].conclusion} does not match theorem {thm.symbols}"

    # root_inf = stack[0]
    # root_inf.print(thm.label)





"""
data structure:

    label -> floating | essential | axiom | proposition
    floating: (label, symbol, type)
    essential: (label, symbols)
    axiom: (label, symbols)
    proposition: (label, symbols, hypotheses, proof)

    scopes? constants? variables? proofs?

"""
from collections import namedtuple
import numpy as np
import os

Declaration = namedtuple("Declaration", ("block", "tag", "symbols"))
Hypothesis = namedtuple("Hypothesis", ("block", "label", "tag", "symbols"))
Axiom = namedtuple("Axiom", ("block", "label", "tag", "symbols"))
Proposition = namedtuple("Proposition", ("block", "label", "tag", "symbols", "proof"))
Scope = namedtuple("Scope", ("c", "v", "d", "f", "e"))

class Rule:
    def __init__(self, conclusion_label, hypothesis_labels):
        self.conclusion = conclusion_label
        self.hypotheses = hypothesis_labels

class Block:
    def __init__(self, parent):
        self.parent = parent
        self.children = []

    def __repr__(self):
        return "${...$}"

    def parent_block(self):
        return self.parent

    def add_sub_block(self):
        self.children.append(Block(parent = self))
        return self.children[-1]

    def add_statement(self, tag, label=None):

        if tag in ("$c", "$v", "$d"):
            statement = Declaration(self, tag, [])
        elif tag == "$p":
            statement = Proposition(self, label, tag, [], [])
        elif tag == "$a":
            statement = Axiom(self, label, tag, [])
        else:
            statement = Hypothesis(self, label, tag, [])

        self.children.append(statement)
        return statement

    def get_scope(self, statement):
        if self.parent == None:
            scope = Scope([], [], [], [], [])
        else:
            scope = self.parent.get_scope(self)

        for child in self.children:
            if child is statement: break
            if type(child) == Block: continue
            if child.tag == "$c": scope.c.extend(child.symbols)
            if child.tag == "$v": scope.v.extend(child.symbols)
            if child.tag == "$f": scope.f.append(child)
            if child.tag == "$e": scope.e.append(child)

        return scope

    def print(self, prefix):
        for child in self.children:
            if type(child) == Block:
                print(prefix + "${")
                child.print(prefix + " ")
                print(prefix + "$}")

            if type(child) == Declaration:
                print(f"{prefix}{child.tag} {' '.join(child.symbols)} $.")
    
            if type(child) == Hypothesis:
                print(f"{prefix}{child.label} {child.tag} {' '.join(child.symbols)} $.")
    
            if type(child) == Proposition:
                print(f"{prefix}{child.label} {child.tag} {' '.join(child.symbols)} $= {' '.join(child.proof)} $.")

class Database:
    def __init__(self):
        self.root_block = Block(parent=None)
        self.statements = {} # looks up statements by label
        self.rules = {} # looks up rules by conclusion's label
    def get_statement(self, label):
        return self.statements[label]
    def get_rule(self, label):
        return self.rules[label]
    def print(self):
        self.root_block.print(prefix = "")

def parse(fpath):

    db = Database()
    block = db.root_block # current database block

    # parser state
    in_comment = False # True if currently in a comment
    in_symbol_list = False # True if currently in a symbol list
    in_proof = False # True if currently in a proof
    label = None # most recent label
    statement = None # most recent statement
    hypotheses = [[]] # stack of hypotheses in current scope

    with open(fpath, "r") as f:
        for n, line in enumerate(f):
            for token in line.strip().split():

                # comments
                if token == "$(": in_comment = True
                if token == "$)": in_comment = False
                if in_comment: continue

                # end of symbol lists or proofs
                if token == "$.":
                    in_symbol_list = in_proof = False

                # scope
                if token == "${":
                    block = block.add_sub_block()
                    hypotheses.append([])
                if token == "$}":
                    block = block.parent_block()
                    hypotheses.pop()

                # declarations
                if token in ("$c", "$v", "$d"):
                    statement = block.add_statement(tag=token, label=None)
                    in_symbol_list = True

                # labeled statements
                if token in ("$f", "$e", "$a", "$p"):
                    assert label != None, \
                           f"line {n+1}: {token} not preceded by label"

                    statement = block.add_statement(tag=token, label=label)
                    db.statements[label] = statement
                    in_symbol_list = True

                    if token in ("$f", "$e"):
                        hypotheses[-1].append(label)
                    else:
                        hyps = [lab for hyps in hypotheses for lab in hyps]
                        db.rules[label] = Rule(label, hyps)

                if token == "$=":
                    in_symbol_list = False
                    in_proof = True
                    assert type(statement) == Proposition, \
                           f"line {n+1}: non-proposition with a proof"

                if token[0] != "$":
                    assert "$" not in token, \
                           f"line {n+1}: '$' not allowed in middle of {token}"

                    if in_symbol_list:
                        statement.symbols.append(token)
                    elif in_proof:
                        statement.proof.append(token)
                    else:
                        label = token

    return db

if __name__ == "__main__":

    fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    # fpath = 'badparse.mm'

    db = parse(fpath)

    # print(db)
    # dbprint(db[-100:])
    db.root_block.children = db.root_block.children[-100:]
    db.print()

    print(f"{len(db.statements)} statements total")
    

    # ### numpy stuff

    # print("loading...")

    # # with open(fpath, "r") as f: raw = f.read()
    # # tokens = raw.split()
    # # line_numbers = []

    # tokens = []
    # line_numbers = []
    # with open(fpath, "r") as f:
    #     for n, line in enumerate(f):
    #         line_tokens = line.strip().split()
    #         tokens.extend(line_tokens)
    #         line_numbers.extend([n]*len(line_tokens))

    # tokens = np.array(tokens)
    # line_numbers = np.array(line_numbers)

    # print(len(tokens))
    # print(len(line_numbers))

    # print("removing comments...")

    # comments = ((tokens == "$(").cumsum() - (tokens == "$)").cumsum()) > 0
    # keep = ~comments & ~(tokens == "$)")    
    # tokens = tokens[keep]
    # line_numbers = line_numbers[keep]

    # print(len(tokens))
    # print(len(line_numbers))

    # print("calculating scope depths...")
    # scope_depth = ((tokens == "${").cumsum() - (tokens == "$}").cumsum())
    # keep = ~(tokens == "${") & ~(tokens == "$}")
    # tokens = tokens[keep]
    # line_numbers = line_numbers[keep]

    # print(len(tokens))
    # print(len(line_numbers))

    # # raw = " ".join(tokens)
    # # assert raw[1:].count("$") == raw.count(" $")

    



# run from mmpy with $ python -m src.mmmine.raw
import sys
import os
import random
import itertools as it
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms

# proof length and size (uncompressed)
def proof_metrics(root):
    size = len(root.conclusion)
    leng = 1
    for dep in root.dependencies.values():
        dsize, dleng = proof_metrics(dep)
        size += dsize
        leng += dleng
    return size, leng

if __name__ == "__main__":

    import matplotlib.pyplot as pt
    from matplotlib import rcParams
    rcParams["font.family"] = "serif"

    db = ms.load_imp()
    # db = ms.load_ni()
    # db = ms.load_pl()
    # db.print()

    # get axioms, split into wff and |-
    axioms = {"wff": {}, "|-": {}, "all": {}}
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$a":
            axioms[rule.consequent.tokens[0]][label] = rule
            axioms["all"][label] = rule
    ax_labels = tuple(axioms["all"].keys())

    # get all wff floating labels in standardized metamath order
    wff_labels = {} # label: rule
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$f" and rule.consequent.tokens[0] == "wff":
            # maintain original sort order
            if label not in wff_labels: wff_labels[label] = 1
    print(wff_labels.keys())

    # get all wff variable labels actually appearing in proofs, in case different
    wff_vars = {} # label: rule
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$p":
            root, _ = mp.verify_proof(db, rule)
            labels = set(wff_labels.keys()) & set(root.normal_proof())
            for label in labels:
                wff_vars[label] = db.rules[label]

    # sort wff_vars to standardized metamath order and discard unused floating labels
    wff_labels = tuple(label for label in wff_labels.keys() if label in wff_vars)
    wff_vars = {label: wff_vars[label] for label in wff_labels}

    print("axioms: " + " ".join(axioms["all"].keys()))
    print("metavars: " + " ".join(wff_labels))

    # all valid proofs (ending with singleton stacks)
    valid_proofs = {} # [proof string] = (proof step, proof length)

    # organize proof stacks by proof size
    stacks = {0: {(): []}} # [size][proof string] = stack

    for size in range(30):

        # skip sizes with no proofs
        if size not in stacks: continue

        print(f"size {size}: {len(valid_proofs)} valid proofs so far, {len(stacks[size])} frontier stacks")
        print(", ".join(f"{sz}:{len(stacks[sz])}" for sz in sorted(stacks.keys())))

        # try appending to all stacks of current size
        for proof, stack in stacks[size].items():

            # try every axiom
            try_labels = ax_labels

            # can also push a new metavariable floating hypothesis, in lexicographic order
            fl_idx = len(set(proof) & set(wff_labels))
            try_labels += wff_labels[:fl_idx+1]

            # try applying each rule
            for label in try_labels:

                new_proof = proof + (label,)
                new_stack = list(stack) # modified in-place
                new_step, msg = mp.conduct(db.rules[label], new_stack)

                # skip if invalid
                if msg != "": continue

                # replace top of stack with existing proof_step if trail is a valid proof
                for start in reversed(range(len(proof))):
                    trail = new_proof[start:]
                    if trail in valid_proofs:
                        (new_step, _) = valid_proofs[trail]
                        break

                # calculate the new proof size
                new_size = size + len(new_step.conclusion)
                if new_size not in stacks:
                    stacks[new_size] = {}

                # push the new step and save the new stack
                new_stack.append(new_step)
                stacks[new_size][new_proof] = new_stack

                # if new stack is singleton, new proof is valid; save it
                if len(new_stack) == 1:
                    valid_proofs[new_proof] = (new_step, new_size)
                    # print("new valid proof:", " ".join(new_proof), " ".join(new_step.conclusion))
   


    # # this method is too coarse iterations; itr 4 takes too long with too many combos

    # # start constructing raw proofs for wff and |=
    # proofs = {"wff": {}, "|-": {}} # proof string: proof_step
    # fails = set() # invalid proof strings

    # # initialize with hypothesis-less proofs (floating hypotheses)
    # proofs["wff"][wff_labels[:1]], msg = mp.perform(db.rules[wff_labels[0]], ())
    # assert msg == ""
    # print(proofs)

    # for itr in range(4):

    #     print(f"\n *** itr {itr}\n")

    #     # apply inferences to existing proofs
    #     new_proofs = {"wff": {}, "|-": {}}
    
    #     for key in axioms:
    #         for label, rule in axioms[key].items():
    #             wff_it = it.product(proofs["wff"], repeat=len(rule.floatings))
    #             ent_it = it.product(proofs["|-"], repeat=len(rule.essentials))


    #             # ! ONCe you commit to the wffs, just substitute into the essentials and they are already determined.
    #             # if you dict all conclusions so far, you can lookup immediately if essential has been proved, skip if not.
    
    #             for wff_proofs, ent_proofs in it.product(wff_it, ent_it):
    
    #                 # lexo substitutions
    
    #                 full_proof = tuple(lab for wff_proof in wff_proofs for lab in wff_proof)
    #                 full_proof += tuple(lab for ent_proof in ent_proofs for lab in ent_proof)
    #                 full_proof += (label,)
    
    #                 if full_proof in proofs[key]: continue
    #                 if full_proof in new_proofs[key]: continue
    #                 if full_proof in fails: continue
    
    #                 deps = tuple(proofs["wff"][wff_proof] for wff_proof in wff_proofs)
    #                 deps += tuple(proofs["|-"][ent_proof] for ent_proof in ent_proofs)
    #                 new_step, msg = mp.perform(rule, deps)
    
    #                 if msg != "":
    #                     fails.add(full_proof)
    #                 else:
    #                     new_proofs[key][full_proof] = new_step
    
    #     for key in proofs: proofs[key].update(new_proofs[key])
    #     # print(proofs)
    #     print(f"\n{len(proofs['wff'])} wff proofs, {len(proofs['|-'])} |- proofs")

    # concs = set()
    # sizes = {}
    # lengs = {}
    # for p, (proof, step) in enumerate(proofs["|-"].items()):
    #     print(p, " ".join(proof), " ".join(step.conclusion))
    #     if " ".join(step.conclusion) == "( ph -> ph )": input('!')
    #     concs.add(step.conclusion)
    #     sizes[step.conclusion], lengs[step.conclusion] = proof_metrics(step)

    # print(f"{len(concs)} distinct conclusions, {len(proofs['|-'])} distinct proofs, {len(fails)} invalid proofs")
    # print(f"max size, leng = {max(sizes.values())}, {max(lengs.values())}")






"""
run from mmpy with $ python -m src.mmmine.wff_repeats

Plot wff length vs frequency across all assertions and proof steps
"normalize" variable names
color-code the ones that are already definitions
ideally, parse wffs and include sub-wffs
"""

import sys
import os
import random
import pickle as pk
import itertools as it
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms

# parse ( definiendum <-> definiens )
# return definiendum, definiens
def parse_def(tokens):

    depth = 0
    for t, token in enumerate(tokens):
        # track parenthesis depth
        if token == "(": depth += 1
        if token == ")": depth -= 1

        # split on shallowest biconditional
        if token == "<->" and depth == 1:
            return tokens[1:t-1], tokens[t+1:-1]

    # parse error if not returned yet
    return None, None

# approximate parentheses-based parse
def parse_wff(tokens):
    sub_wffs = set([tokens])
    open_idx = {}
    depth = 0
    for t, token in enumerate(tokens):
        # track parenthesis depth
        if token == "(":
            depth += 1
            open_idx[depth] = t
        if token == ")":
            sub_wff = tokens[open_idx[depth]:t+1]
            sub_wffs.add(sub_wff)
            depth -= 1
    return sub_wffs    

canon_t = ("ph","ps","ch","th","ta","et","ze","si","rh","mu","la","ka")
canon_s = set(canon_t)

def normalize(tokens):
    subst = {}
    for t, token in enumerate(tokens):
        if token not in canon_s: continue
        if token in subst: continue
        subst[token] = canon_t[len(subst)]
    return tuple(subst.get(tok, tok) for tok in tokens)

if __name__ == "__main__":

    do_count = False

    if do_count:

        db = ms.load_pl()
        # db = ms.load_all()

        wff_counts = {}
        def_wffs = set()
        parse_errors = []

        for r, (label, rule) in enumerate(db.rules.items()):
            print(f"rule {r} ({label}) of {len(db.rules)}...")

            # include consequent wff
            wff = tuple(rule.consequent.tokens)
            wff = wff[1:] # exclude leading wff or |-
            sub_wffs = parse_wff(wff)
            for sub_wff in map(normalize, sub_wffs):
                wff_counts[sub_wff] = wff_counts.get(sub_wff, 0) + 1

            # mark wffs that are already definitions
            if rule.consequent.tag == "$a" and label[:3] == "df-":
                definiendum, definiens = parse_def(wff)
                if definiens is None:
                    parse_errors.append(wff)
                else:
                    definiens = normalize(definiens)
                    wff_counts[definiens] = wff_counts.get(definiens, 0) + 1
                    def_wffs.add(definiens)

            # only look at proofs of propositions
            if rule.consequent.tag != "$p": continue

            # uncompress proof
            root, steps = mp.verify_proof(db, rule)

            # count wffs in each step
            for step in steps.values():
                wff = tuple(step.conclusion)
                wff = wff[1:] # exclude leading wff or |-
                sub_wffs = parse_wff(wff)
                for sub_wff in map(normalize, sub_wffs):
                    wff_counts[sub_wff] = wff_counts.get(sub_wff, 0) + 1
    
        with open("wff_repeats.pkl", "wb") as f:
            pk.dump((wff_counts, def_wffs, parse_errors), f)

    with open("wff_repeats.pkl", "rb") as f:
        wff_counts, def_wffs, parse_errors = pk.load(f)

    print(f"{len(wff_counts)} wffs, {len(def_wffs)} are defs")
    print(f"{len(parse_errors)} parse errors:")
    for tokens in parse_errors: print(" ".join(tokens))


    wffs, counts = zip(*wff_counts.items())
    lens = np.array(list(map(len, wffs)))
    counts = np.array(counts)
    is_def = np.array([wff in def_wffs for wff in wffs])

    sorter = np.argsort(-counts)
    wffs = tuple(wffs[i] for i in sorter)
    lens = lens[sorter]
    counts = counts[sorter]
    is_def = is_def[sorter]

    dominators = np.zeros(len(counts))
    for i in range(len(counts)):
        dominators[i] = ((lens >= lens[i]) & (counts >= counts[i])).sum() - 1

    print("non-dominated non-def wffs:")
    for idx in np.flatnonzero((dominators == 0) & ~is_def):
        print(f"{counts[idx]:4d}: {' '.join(wffs[idx])[:50]}")

    # print("top most frequent wffs:")
    # sorter = np.argsort(-counts)
    # for idx in sorter[:30]:
    #     print(f"{counts[idx]: 4d}: {' '.join(wffs[idx])}")

    import matplotlib.pyplot as pt
    from matplotlib import rcParams
    rcParams["font.family"] = "serif"

    pt.figure(figsize=(3.75,3))
    pt.plot(lens[is_def], counts[is_def], mfc='w', mec='k', marker='o', markersize=10, linestyle='none', label="Definition", zorder=2)
    pt.plot(lens[(dominators == 0) & ~is_def], counts[(dominators == 0) & ~is_def], 'k+', label="Pareto optimal", zorder=1)
    pt.plot(lens[(dominators >= 1) & ~is_def], counts[(dominators >= 1) & ~is_def], 'k.', label="other", zorder=0)
    pt.xlabel("Length")
    pt.ylabel("Frequency")
    pt.xscale("log")
    pt.yscale("log")
    pt.legend()
    pt.tight_layout()
    pt.savefig("wff_repeats.eps")
    pt.show()




from time import perf_counter
import mmmine.bidao as bd

if __name__ == "__main__":

    import metamathpy.setmm as ms

    # individual targets

    # # target = "bijust" # justification theorems
    # target = "meredith" # axiom equivalence
    # num_sweeps = 2

    # db = ms.load_to(target)

    # goal_tokens = tuple(db.rules[target].consequent.tokens)

    # # just keep metavariables in floating hypotheses for now
    # metavars = tuple(hyp.tokens[1] for hyp in db.rules[target].floatings)

    # # filter rules
    # rules = bd.filter_rules(db, metavars, ax_only=False, dependent_floor=2, exclude_labels=(target,))
    # # for rule in rules[:20]: print(rule)
    # print(f"{len(db.rules)} to {len(rules)} filtered rules")

    # # run sweeps
    # root_step = bd.run_sweeps(num_sweeps, rules, goal_tokens, dbg=True)

    # lots of targets
    num_sweeps = 3
    max_time = 5

    db = ms.load_pl()

    prop_labels = []
    for rule in db.rules.values():
        if rule.consequent.tag == "$p" and len(rule.essentials)==0: prop_labels.append(rule.consequent.label)

    num_success = 0
    for t, target in enumerate(prop_labels):

        goal_tokens = tuple(db.rules[target].consequent.tokens)
        print(f"Target {t} of {len(prop_labels)}, {num_success} success so far: {' '.join(goal_tokens)}")

        # just keep metavariables in floating hypotheses for now
        metavars = tuple(hyp.tokens[1] for hyp in db.rules[target].floatings)
    
        # filter rules
        rules = bd.filter_rules(db, metavars, ax_only=False, dependent_floor=1, exclude_labels=prop_labels[t:])
        # for rule in rules[:20]: print(rule)
        print(f"{len(db.rules)} to {len(rules)} filtered rules")
    
        # run sweeps
        timeout = perf_counter() + max_time
        root_step = bd.run_sweeps(num_sweeps, rules, goal_tokens, timeout, dbg=False)
        if root_step is not None: num_success += 1
    



# run from mmpy with $ python -m src.mmmine.used
import sys
import os
import random
import pickle as pk
import itertools as it
import numpy as np
from time import perf_counter
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms

def sample_formulation(db, sample_size, verbose=False):

    # collect all the theorem labels
    thm_labels = set()
    for label, rule in db.rules.items():
        if rule.consequent.tag == "$p":
            thm_labels.add(label)

    if verbose: print(f"{len(thm_labels)} theorems total")

    # sub-sample some theorem labels
    sample_labels = set(random.sample(list(thm_labels), sample_size))
    if verbose: print(sample_labels)

    # populate all labels of formulation
    formulation_labels = set()

    # recursive helper to add all other rules used in sample's proofs
    # modifies formulation_labels in place
    def add_all(label):

        rule = db.rules[label]

        # if label already included, don't repeat
        if label in formulation_labels: return

        # if label is essential hypothesis, don't include as valid inference rule
        if rule.consequent.tag == "$e": return

        # include current label if it is axiom, floating hypothesis, or proposition
        formulation_labels.add(label)

        # only recurse on labels of propositions
        if rule.consequent.tag != "$p": return

        # expand proof to get all other rules it needs
        root, steps = mp.verify_proof(db, rule)

        # recurse on rules used to justify each step
        for step in steps.values():
            add_all(step.rule.consequent.label)

    # use recursive helper on every sampled label
    for i,label in enumerate(sample_labels):
        if verbose: print(i, label, len(formulation_labels))
        add_all(label)

    # now formulation labels are populated, return
    return formulation_labels

def formulation_size(db, formulation_labels):
    # tally size over all steps of all proofs
    size = 0
    for label in formulation_labels:
        rule = db.rules[label]

        # only propositions have proof, otherwise skip
        if rule.consequent.tag != "$p": continue 

        # generate proof steps
        root, steps = mp.verify_proof(db, rule)

        # add up step lengths, conclusions are keys
        for conclusion in steps: size += len(conclusion)

    return size

def forward_search(db, formulation_labels, max_depth):

    valid_proofs = {d: {} for d in range(max_depth+1)} # depth -> proof string -> conclusion step
    valid_proofs["all"] = {} # depth|all
    stacks = {0: {(): []}} # depth -> proof string -> stack

    # search by depth
    for depth in range(max_depth):

        # allocate stacks for new depth
        stacks[depth+1] = {}

        # try appending labels to each current stack
        for proof, stack in stacks[depth].items():
            for label in formulation_labels:

                new_proof = proof + (label,)
                new_stack = list(stack) # will be modified in-place
                new_step, msg = mp.conduct(db.rules[label], new_stack)

                # skip if invalid
                if msg != "": continue

                # push the new step and save the new stack
                new_stack.append(new_step)
                stacks[depth+1][new_proof] = new_stack

                # if new stack is singleton, new proof is valid; save it
                if len(new_stack) == 1:
                    valid_proofs["all"][new_proof] = new_step
                    valid_proofs[depth+1][new_proof] = new_step
                    # print("new valid proof:", " ".join(new_proof), " ".join(new_step.conclusion))

    return valid_proofs, stacks

if __name__ == "__main__":

    db = ms.load_pl()
    # db = ms.load_all()

    num_samples = 500
    sample_size = 1
    max_depth = 5

    do_expansion = False

    if do_expansion:

        samples = []
        membranes = [] # new theorem sets after expansion
        growths = [] # per-depth search growth
        for s in range(num_samples):
            formulation_labels = sample_formulation(db, sample_size)
            samples.append(formulation_labels)
            print(f"{s}: {len(formulation_labels)} labels in formulation")
    
            # get conclusions of existing formulation labels
            existing = set()
            for label in formulation_labels:
                existing.add(tuple(db.rules[label].consequent.tokens))
    
            # try expanding
            valid_proofs, stacks = forward_search(db, formulation_labels, max_depth)
            for depth in range(max_depth+1):
                print(f"  depth {depth}: {len(valid_proofs[depth])} valid proofs, {len(stacks[depth])} stacks")
    
            # count distinct conclusions separate from what is already in formulation
            conclusions = {} # conclusion -> [proof strings]
            for proof, final_step in valid_proofs["all"].items():
                conc = final_step.conclusion
                if conc not in conclusions: conclusions[conc] = []
                conclusions[conc].append(proof)
    
            new_theorems = {k: v for (k,v) in conclusions.items() if k not in existing}
    
            print(f" {len(valid_proofs['all'])} valid proofs, {len(conclusions)} distinct conclusions, {len(new_theorems)} new")

            membranes.append((conclusions, new_theorems))
            growths.append([len(valid_proofs[d]) for d in range(max_depth+1)])

        with open("expansion.pkl", "wb") as f:
            pk.dump((samples, membranes, growths), f)

    with open("expansion.pkl", "rb") as f:
        (samples, membranes, growths) = pk.load(f)

    form_sizes = [formulation_size(db, sample) for sample in samples]
    thm_counts = [len(new_theorems) for (_, new_theorems) in membranes]
    dist_concs = [len(conclusions) for (conclusions, _) in membranes]
    growths = np.array(growths)


    import matplotlib.pyplot as pt
    from matplotlib import rcParams
    rcParams["font.family"] = "serif"

    pt.figure(figsize=(8,2.5))
    pt.subplot(1,2,1)
    pt.hist(list(map(len, samples)), rwidth=.8)
    pt.xlabel("Rules in formulation")
    pt.ylabel("Frequency")

    pt.subplot(1,2,2)
    pt.hist(form_sizes, rwidth=.8)
    pt.xlabel("Formulation size")
    # pt.ylabel("Frequency")

    pt.tight_layout()
    pt.savefig("formsamples.eps")

    pt.figure(figsize=(8,2.5))
    pt.subplot(1,2,1)
    pt.plot(growths.T, 'k-', alpha=.05)
    pt.plot(growths.mean(axis=0), 'b-')
    pt.xlabel("Search depth")
    pt.ylabel("Valid proofs")
    pt.yscale('log')

    # pt.subplot(1,3,2)
    # pt.plot(dist_concs, growths.sum(axis=1), 'k.')
    # pt.xlabel("Distinct conclusions")
    # pt.ylabel("Valid proofs")

    pt.subplot(1,2,2)
    pt.plot(form_sizes, thm_counts, 'k.')
    pt.xlabel("Formulation size")
    pt.ylabel(f"{max_depth}-expansion")

    pt.tight_layout()
    pt.savefig("expansion.pdf")

    pt.show()




"""
Run from top-level directory with
$ python -m src.metamathpy.environment
"""
from metamath.proof import conduct

"""
Environment steps through normal proof verification process
Resets with a new claim to be proved and empty stack
State is current (claim, proof, stack)
Action is rule label
Step returns state, error message ("" means valid)
Done when top of stack matches claim

Attributes:
env.claim is current claim being proved
env.stack is the stack
env.proof is the list of labels in the uncompressed proof
"""
class Environment:
    def __init__(self, database):
        self.db = database
        self.claim = None
        self.proof = []
        self.stack = []
        self.steps = []

    def copy(self):
        env = Environment(self.db)
        env.claim = self.claim
        env.proof = list(self.proof)
        env.stack = list(self.stack)
        env.steps = list(self.steps)
        return env

    def reset(self, claim=None):
        self.claim = claim
        self.proof = []
        self.stack = []
        self.steps = []
        return (self.claim, list(self.proof), list(self.stack))

    # action is label of rule to apply or integer index
    # index will push a previous proof step back onto stack
    def step(self, action):

        # enforce disjoint variables if proving a specific claim
        disjoint = set() if self.claim is None else self.claim.disjoint

        # conduct next proof step
        if type(action) is str:

            rule = self.db.rules[action]
            proof_step, msg = conduct(rule, self.stack, disjoint)

        else:

            if type(action) is int and 0 <= action < len(self.steps):
                proof_step, msg = self.steps[action], ""
            else:
                proof_step, msg = None, f"{action} out of bounds for length {len(self.steps)} proof"

        # push proof step onto stack
        self.stack.append(proof_step)

        # save actions so far in proof
        self.proof.append(action)

        # save steps so far in proof
        self.steps.append(proof_step)

        # return state and message
        state = (self.claim, list(self.proof), list(self.stack))
        return state, msg

if __name__ == "__main__":

    from metamath.database import *
    import os
    fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = parse(fpath)
    print("parsed.")

    env = Environment(db)

    print("mpd proof:")
    claim = db.rules["mpd"]
    proof = ['wph', 'wph', 'wph', 'wi', 'wph', 'wph', 'wph', 'ax-1', 'wph', 'wph', 'wph', 'wi', 'ax-1', 'mpd']
    env.reset(claim)

    for a, action in enumerate(proof):
        (claim, _, stack), msg = env.step(action)
        assert msg == ""
        print(f"{a}: action = {action}")
        print('top of stack', stack[-1])

    print("\ninvalid step proof:")
    proof = ["wph", "wi"]
    env.reset()
    for a, action in enumerate(proof):
        (claim, _, stack), msg = env.step(action)
        print(f"{a}: action = {action}")
        print('top of stack:', stack[-1])
    assert msg != ""
    print(f"error message: {msg}")

    print("\nopen-ended:")
    proof = ["wph", "wps", "ax-1", "wph"]
    env.reset()
    for a, action in enumerate(proof):
        (claim, _, stack), msg = env.step(action)
        print(f"{a}: action = {action}, stack:")
        print(stack)

    env = Environment(db)
    print(env.reset())
    
    print(env.step("wph"))
    env2 = env.copy()

    print(env.step("wi"))
    print(env2.step("wps"))







import os
# from ..metamathpy import database as md
from metamathpy import database as md

def load_imp(fpath = None):
    # last label before any ax-3 proofs is loowoz
    print('loading..')
    if fpath is None:
        fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = md.parse(fpath, last_rule="loowoz")
    # negation wffs and ax-3 not used
    db.rules.pop("wn")
    db.rules.pop("ax-3")
    return db

def load_ni(fpath = None):
    # last label before any new boolean operator definitions is bijust, "rule" 441
    print('loading..')
    if fpath is None:
        fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = md.parse(fpath, last_rule="bijust")
    return db

def load_pl(fpath = None):

    # last label before any FOL (universal quantifier) is xorexmid, it is "rule" 2849 (including hypotheses)
    print('loading..')
    if fpath is None:
        fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = md.parse(fpath, last_rule="xorexmid")
    return db

def load_all(fpath = None):
    print('loading..')
    if fpath is None:
        fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = md.parse(fpath)
    return db

def load_to(last_rule, fpath = None):
    print('loading..')
    if fpath is None:
        fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    db = md.parse(fpath, last_rule=last_rule)
    return db

if __name__ == "__main__":

    # db = load_imp()
    # db = load_ni()
    db = load_pl()
    db.print()




import os

def remove_comments(raw):

    # separate into chunks starting with comment text (except first)
    split = raw.split("$(")
    line = split[0].count(os.linesep) + 1 # track line numbers for error message
    
    # remove leading comment text from each chunk
    for i in range(1, len(split)):
    
        # get endpoint of leading comment, crash if unbalanced
        close = split[i].find("$)")
        assert close > -1, f"unbalanced '$(' starting on line {line}"
    
        # update line count and remove leading comment
        line += split[i].count(os.linesep)
        split[i] = split[i][close+2:]
    
    # join back into text with comments removed
    return "".join(split)

fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
# fpath = 'badparse.mm'

with open(fpath, "r") as f: raw = f.read()

# # ### with re.sub?
# import re
# commentless = re.sub(r"\$\(.+?\$\)", "", raw)

commentless = remove_comments(raw)

tokens = commentless.split()

print(tokens[:20])
        






from time import perf_counter
import itertools as it
import numpy as np
import metamathpy.proof as mp
import metamathpy.database as md
from metamathpy.unification import substitute

# place-holder rule for wff parses
WFF_RULE = md.Rule(md.Statement("WFF", None, None, None), [], [], set(), set())

class GoalNode:
    def __init__(self, conclusion):
        # conclusion: token string to be proved
        self.conclusion = conclusion

        # each child represents an inference that would prove this goal if essential dependencies are satisfied
        # each essential dependency of a child has a respective sub-goal node
        # assumes floating dependencies already proven
        # successful proof requires logical-or over the children, logical-and of their essential depency token strings
        self.children = {} # tuple of essential dependency token strings -> rule, substitution, floating dependencies, sub-goals

    def update(self, rule, substitution, floating_steps, essential_tokens):
        # save a new route to prove self (no-op if route already included in children)
        # floating_steps[n] is proof step satisfying nth floating dependency
        # essential_tokens[n] are tokens for the nth essential dependency sub-goal
        # each essential dependency gets a sub-goal node
        # returns sub-goals for this update (does not check for duplicates or cycles in the DAG)

        # don't update a node with an existing route
        if essential_tokens in self.children: return ()

        # otherwise construct requisite sub-goals
        sub_goals = tuple(map(GoalNode, essential_tokens))

        # update with new route and return
        self.children[essential_tokens] = (rule, substitution, floating_steps, sub_goals)
        return sub_goals

    def satisfy_from(self, proof_steps):
        # proof_steps[conclusion tokens] = first proof step encountered with that conclusion (type |- only), as in rule sweep
        # try to satisfy goal from proof steps, returning None if failed or the final proof step if successful

        # base case
        if self.conclusion in proof_steps:
            return proof_steps[self.conclusion]

        # recursive case: logical-or over children
        for essential_tokens in self.children:

            (rule, substitution, floating_steps, sub_goals) = self.children[essential_tokens]

            # logical-and over child sub-goals: any None sub_proofs indicate not satisfied
            sub_proofs = tuple(sub_goal.satisfy_from(proof_steps) for sub_goal in sub_goals)
            if None in sub_proofs: continue

            # if this line is reached, all sub-goals were proved
            proof_step, status = mp.perform(rule, floating_steps + sub_proofs)
            assert status == "" # make sure inference is valid
            return proof_step

        # if this line reached, logical-or failed
        return None

    def deficit(self, proof_steps):
        # proof_steps[conclusion tokens] = first proof step encountered with that conclusion (type |- only), as in rule sweep
        # count the minimal number of open descendent goals needed to satisfy self
        # (todo: avoid double-counting duplicates)
        
        # base case: self already satisfied directly
        if self.conclusion in proof_steps: return 0

        # base case: no children, so self must be satisfied directly
        if len(self.children) == 0: return 1

        # recursive case: logical-or (minimum) over children
        child_min = np.inf
        for essential_tokens in self.children:
        
            # logical-and (sum) over current child's sub-goals
            (rule, substitution, floating_steps, sub_goals) = self.children[essential_tokens]
            child_sum = sum(sub_goal.deficit(proof_steps) for sub_goal in sub_goals)
            child_min = min(child_min, child_sum)

        return child_min

    def tree_string(self, prefix=""):
        s = prefix + " ".join(self.conclusion) + " [OR:\n"
        for essential_tokens in self.children:
            (rule, substitution, floating_steps, sub_goals) = self.children[essential_tokens]
            s += prefix + f" {rule.consequent.label}[AND:\n"
            for sub_goal in sub_goals:
                s += sub_goal.tree_string(prefix + "  ")
            s += prefix + " :DNA]\n"
        s += prefix + ":RO]\n"
        return s


def rule_sweep(rules, proof_steps, goals, timeout):
    # rules[n] = nth inference rule to try
    # proof_steps[typecode][conclusion tokens] = first proof step encountered with that type code and conclusion
    # goals[conclusion tokens][k] = kth encountered goal node for that conclusion (todo: dag but with cycle-checks)
    # timeout: once perf_counter exceeds this, ends early
    new_steps = {"wff": {}, "|-": {}} # other typecodes needed beyond PL
    new_goals = {}
    for rule in rules:

        # early termination
        if perf_counter() > timeout: break

        # apply rule to all wff combinations for metavariables
        for wffs in it.product(proof_steps["wff"].values(), repeat=len(rule.floatings)):

            # early termination
            if perf_counter() > timeout: break

            # setup substitution from floating hypotheses
            substitution = {hyp.tokens[1]: dep.conclusion[1:] for (hyp, dep) in zip(rule.floatings, wffs)}

            # apply substitution to essential hypotheses and consequent
            essential_tokens = tuple(substitute(hyp.tokens, substitution) for hyp in rule.essentials)
            conclusion_tokens = substitute(rule.consequent.tokens, substitution)

            # forward chaining: only apply rule if hypotheses are satisfied
            if all(tokens in proof_steps["|-"] for tokens in essential_tokens):

                # set up all dependencies
                deps = wffs + tuple(proof_steps["|-"][tokens] for tokens in essential_tokens)
    
                # perform the rule
                step, status = mp.perform(rule, deps)
    
                # should be sound for PL (no distinct variable requirements)
                assert status == ""
    
                # save new results
                tokens = step.conclusion
                typecode = tokens[0]
                if tokens not in proof_steps[typecode] and tokens not in new_steps[typecode]:
                    new_steps[typecode][tokens] = step
    
                # also assume any conclusion is well-formed
                if typecode == "|-":
                    tokens = ("wff",) + tokens[1:] 
                    if tokens not in proof_steps["wff"] and tokens not in new_steps["wff"]:
                        new_steps["wff"][tokens] = mp.ProofStep(tokens, rule=WFF_RULE) # skip the actual parse

            # backward chaining: update and-or tree if goal is matched
            if conclusion_tokens in goals:
                for goal in goals[conclusion_tokens]:
                    sub_goals = goal.update(rule, substitution, wffs, essential_tokens)
                    for sub_goal in sub_goals:
                        if sub_goal.conclusion not in new_goals:
                            new_goals[sub_goal.conclusion] = []
                        new_goals[sub_goal.conclusion].append(sub_goal)

            # print(substitution)
            # input('.')

    return new_steps, new_goals

def run_sweeps(num_sweeps, rules, goal_tokens, timeout, dbg=False):
    # todo: multiple top-level goals
    # returns root proof step if successful else None

    # set up goals
    root_goal = GoalNode(goal_tokens)
    goals = {goal_tokens: [root_goal]}

    proof_steps = {"wff": {}, "|-": {}}
    for sweep in range(num_sweeps):

        # do the next sweep
        start = perf_counter()
        new_steps, new_goals = rule_sweep(rules, proof_steps, goals, timeout)
        duration = perf_counter() - start
    
        if dbg:
            print(f"sweep {sweep}:")
            for typecode in ("wff", "|-"):
                print(f"new {typecode}s ({len(new_steps[typecode])} total):")
                for t, tokens in enumerate(new_steps[typecode]):
                    print(f" {' '.join(tokens)}")
                    if t == 10:
                        print(" ...")
                        break

            print(f"new goals ({len(new_goals)} total):")
            for t, tokens in enumerate(new_goals):
                print(f" {' '.join(tokens)}")
                if t == 10:
                    print(" ...")
                    break

        # integrate new proof steps, without replacing earliest encounters
        for typecode in ("wff", "|-"):
            proof_steps[typecode] = new_steps[typecode] | proof_steps[typecode]

        # integrate new goals (todo: deduplicate)
        for conclusion in new_goals:
            if conclusion not in goals: goals[conclusion] = []
            goals[conclusion].extend(new_goals[conclusion])                

        if dbg:
            print("goal tree:")
            print(root_goal.tree_string())
            print(f"deficit = {root_goal.deficit(proof_steps['|-'])}")
    
            print(f"Took {duration:.3f}s ({duration/60:.3f}m)")

        # check satisfaction
        root_step = root_goal.satisfy_from(proof_steps["|-"])
        if root_step is not None:
            print("Root is satisfied!")
            print(root_step.tree_string())
            return root_step

        if dbg: input("..")

def filter_rules(db, metavars, ax_only, dependent_floor, exclude_labels):
    # filter db rules by given criteria
    # only include given metavars
    # omit essential hypotheses
    # omit $p rules if ax_only = True
    # omit rules with less than dependent_floor direct dependents
    # omit excluded labels

    # calculate num direct dependents of each rule, assuming compressed proofs
    num_dependents = {label: 0 for label in db.rules}

    for rule in db.rules.values():
        # only propositions have proofs
        if rule.consequent.tag != "$p": continue

        # get proof labels (assumes proof in compressed format)
        split = rule.consequent.proof.index(")")
        step_labels = rule.consequent.proof[1:split]

        for label in step_labels:
            num_dependents[label] += 1

    rules = []
    for (label, rule) in db.rules.items():

        # for now, omit "rules" for essential hypotheses
        if rule.consequent.tag == "$e": continue

        # if requested, omit any non-axiom propositions
        if ax_only and rule.consequent.tag == "$p": continue

        # limit metavariables
        if rule.consequent.tag == "$f" and rule.consequent.tokens[1] not in metavars: continue

        # omit propositions with few direct dependents
        if rule.consequent.tag == "$p" and num_dependents[label] < dependent_floor: continue

        # omit excluded labels
        if label in exclude_labels: continue

        # otherwise keep the rule
        rules.append(rule)

    return rules

if __name__ == "__main__":

    import metamathpy.setmm as ms

    num_sweeps = 3
    max_time = 10

    # ax_only = True
    # loader = ms.load_pl
    # goal_tokens = tuple("|- ( ph -> ( ps -> ph ) )".split(" "))
    # metavars = ("ph", "ps")

    ax_only = False
    loader = lambda : ms.load_to("mpd")
    goal_tokens = tuple("|- ( ph -> ph )".split(" "))
    metavars = ("ph",)

    db = loader()

    # filter rules
    rules = filter_rules(db, metavars, ax_only, dependent_floor=0, exclude_labels=())

    # run sweeps
    timeout = perf_counter() + max_time
    root_step = run_sweeps(num_sweeps, rules, goal_tokens, timeout, dbg=True)

    # # set up goals
    # root_goal = GoalNode(goal_tokens)
    # goals = {goal_tokens: [root_goal]}

    # proof_steps = {"wff": {}, "|-": {}}
    # for sweep in range(num_sweeps):

    #     # do the next sweep
    #     start = perf_counter()
    #     new_steps, new_goals = rule_sweep(rules, proof_steps, goals)
    #     duration = perf_counter() - start
    
    #     print(f"sweep {sweep}:")
    #     for typecode in ("wff", "|-"):
    #         print(f"new {typecode}s ({len(new_steps[typecode])} total):")
    #         for t, tokens in enumerate(new_steps[typecode]):
    #             print(f" {' '.join(tokens)}")
    #             if t == 10:
    #                 print(" ...")
    #                 break

    #     print(f"new goals ({len(new_goals)} total):")
    #     for t, tokens in enumerate(new_goals):
    #         print(f" {' '.join(tokens)}")
    #         if t == 10:
    #             print(" ...")
    #             break

    #     # integrate new proof steps, without replacing earliest encounters
    #     for typecode in ("wff", "|-"):
    #         proof_steps[typecode] = new_steps[typecode] | proof_steps[typecode]

    #     # integrate new goals (todo: deduplicate)
    #     for conclusion in new_goals:
    #         if conclusion not in goals: goals[conclusion] = []
    #         goals[conclusion].extend(new_goals[conclusion])                

    #     print("goal tree:")
    #     print(root_goal.tree_string())
    #     print(f"deficit = {root_goal.deficit(proof_steps['|-'])}")

    #     print(f"Took {duration:.3f}s ({duration/60:.3f}m)")

    #     root_step = root_goal.satisfy_from(proof_steps["|-"])
    #     if root_step is not None:
    #         print("Root is satisfied!")
    #         print(root_step.tree_string())
    #         break

    #     input("..")





"""
symbols[n]: nth token in symbol string
substitution[v]: symbol string to put in place of symbol v
returns result[n]: nth token after substitutions applied
"""
# @profile
def substitute(symbols, substitution):
    result = ()
    for symbol in symbols:
        if symbol in substitution: result += substitution[symbol]
        else: result += (symbol,)
    return result
# from substitute import substitute # cython version

# a subform function for set.mm
def subform_set(s):
    pass

# unify essential hypotheses of a rule with a sequence of dependencies
# each dependency is a ProofStep object
# also required function handle subform
#   subform(s) returns leading string s[:n] that is a well-formed formula
def unify(rule, dependencies, subform):

    # need enough dependencies to match essential hypotheses
    if len(dependencies) < len(rule.essentials):
        return False, {}

    # match with dependencies at top of stack
    dependencies = dependencies[len(dependencies)-len(rule.essentials):]

    # concatenate single symbol strings for essentials and dependencies
    hyp_tokens, dep_tokens = (), ()
    for hyp, dep in zip(rule.essentials, dependencies):
        hyp_tokens += hyp.tokens
        dep_tokens += dep.conclusion.tokens

    # standardize apart the rule variables, assuming dependencies have no $n tokens
    standardizer = {v: f"${n}" for n, v in enumerate(rule.variables)}
    hyp_tokens = substitute(hyp_tokens, standardizer)

    while hyp_tokens != dep_tokens:
        pass




# run from mmpy with $ python -m src.mmmine.raw_heuristic
import sys
import os
import random
import itertools as it
import numpy as np
from time import perf_counter
from collections import deque
from ..metamathpy import database as md
from ..metamathpy import proof as mp
from ..metamathpy import setmm as ms

if __name__ == "__main__":

    max_proofs = 500_000
    max_queued = 5_000_000

    # import matplotlib.pyplot as pt
    # from matplotlib import rcParams
    # rcParams["font.family"] = "serif"

    # db = ms.load_imp()
    db = ms.load_ni()

    # get axioms, split into wff and |-
    axioms = {"wff": {}, "|-": {}, "all": {}}
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$a":
            # axioms type specific
            axioms[rule.consequent.tokens[0]][label] = rule
            axioms["all"][label] = rule
    ax_labels = tuple(axioms["all"].keys())

    # get all wff floating labels in standardized metamath order
    wff_labels = {} # label: rule
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$f" and rule.consequent.tokens[0] == "wff":
            # maintain original sort order
            if label not in wff_labels: wff_labels[label] = 1
    print(wff_labels.keys())

    # get all wff variable labels actually appearing in proofs, in case different
    wff_vars = {} # label: rule
    for r, (label, rule) in enumerate(db.rules.items()):
        if rule.consequent.tag == "$p":
            root, _ = mp.verify_proof(db, rule)
            labels = set(wff_labels.keys()) & set(root.normal_proof())
            for label in labels:
                wff_vars[label] = db.rules[label]

    # sort wff_vars to standardized metamath order and discard unused floating labels
    wff_labels = tuple(label for label in wff_labels.keys() if label in wff_vars)
    wff_vars = {label: wff_vars[label] for label in wff_labels}

    print("axioms: " + " ".join(f"{label}[{len(axioms['all'][label].hypotheses)}]" for label in axioms["all"].keys()))
    print("metavars: " + " ".join(wff_labels))
    input("Enter to start search...")

    # set up queues of partial proofs
    # queues[k][n] = (label sequence, stack) for nth partial proof of priority k
    # priority is label sequence length (L) + heuristic (x)
    # x = ceil( (S-1) / (H-1) ), where S is stack length and H is the maximum hypothesis count for any axiom
    # this ensures "admissibility" (at least x more inference steps will be needed for a complete proof)
    queues = {0: deque([((), [])])}
    H = max(len(rule.hypotheses) for rule in axioms["all"].values())
    num_queued = 1

    # set up storage of complete proofs and proved statements
    # complete_proofs[label_sequence] = conclusion proof step
    complete_proofs = {}
    proven_statements = {}

    # start queue-based proof search
    while True:
        # stopping criteria
        if len(complete_proofs) >= max_proofs: break
        if num_queued >= max_queued: break

        # pop next partial proof (infinitely many proofs so there will always be at least one)
        k = min([k for k,v in queues.items() if len(v)>0])
        label_sequence, stack = queues[k].popleft()

        # try applying every axiom
        try_labels = ax_labels

        # can also push a new metavariable floating hypothesis, in lexicographic order
        fl_idx = len(set(label_sequence) & set(wff_labels))
        try_labels += wff_labels[:fl_idx+1]

        # try applying each rule
        for label in try_labels:

            new_label_sequence = label_sequence + (label,)
            new_stack = list(stack) # modified in-place
            new_step, msg = mp.conduct(db.rules[label], new_stack)

            # skip if invalid
            if msg != "": continue

            # push the new step
            new_stack.append(new_step)

            # calculate priority for new partial proof
            L = len(new_label_sequence)
            S = len(new_stack)
            x = int(np.ceil( (S-1) / (H-1) ))
            k = L + x
            if k not in queues: queues[k] = deque()

            # update the search queues
            new_entry = (new_label_sequence, new_stack)
            queues[k].append(new_entry)
            num_queued += 1

            # if new stack is singleton, new proof is valid; save it along with the proven statement
            if len(new_stack) == 1:
                complete_proofs[new_label_sequence] = new_step
                conclusion = ' '.join(new_step.conclusion)

                if conclusion not in proven_statements:
                    proven_statements[conclusion] = []
                proven_statements[conclusion].append(new_label_sequence)

                print(f"proof {len(complete_proofs)} [{num_queued} queued, {len(proven_statements)} proven, k<={max(queues.keys())}]: {conclusion} $= {' '.join(new_label_sequence)}")

    input("Search terminated, Enter to list proven theorems...")
    for t, (conclusion, proofs) in enumerate(proven_statements.items()):
        print(f"{t} [{len(proofs)} proofs]: {conclusion}:")
        for p, proof in enumerate(proofs): print(f" $= {' '.join(proof)}")




"""
Run from parent directory with
$ python -m src.database
"""
from collections import namedtuple
import itertools as it

try:
    profile
except NameError:
    profile = lambda x: x

"""
Statement
    label: the label of the statement (str)
    tag: the type of statement ("$p", "$a", etc., str)
    tokens[n]: the nth token in the statement math symbol string (str)
    proof[n]: the nth token in the proof (str)
Rule:
    consequent: the rule's conclusion statement
    essentials[n]: the nth essential hypothesis statement
    floatings[n]: the nth floating hypothesis statement
Frame:
    frame[tag][n]: the nth
        constant or variable symbol if tag in "cv"
        list of disjoint variables if tag is "d"
        hypothesis if tag in "ef"
"""
Statement = namedtuple('Statement', ('label', 'tag', 'tokens', 'proof'))

class Rule:
    def __init__(self, consequent, essentials, floatings, disjoint, variables):
        self.consequent = consequent
        self.essentials = essentials
        self.floatings = floatings
        self.disjoint = disjoint
        self.variables = variables
    def finalize(self):
        self.hypotheses = self.floatings + self.essentials
    def __str__(self):
        s = f"{self.consequent.label} {self.consequent.tag} {' '.join(self.consequent.tokens)} $.\n"
        s += f"disjoint variable sets: {self.disjoint}\n"
        s += f"hypotheses:\n"
        for hypothesis in self.hypotheses:
            s += f"  {hypothesis.label} {hypothesis.tag} {' '.join(hypothesis.tokens)} $.\n"
        return s

def new_frame(): return {tag: [] for tag in "cvdfe"}

class Database:
    def __init__(self):
        self.statements = {} # looks up statements by label
        self.rules = {} # looks up rules by consequent's label

    def print(self, start=0, stop=0):
        if start < 0: start = len(self.rules) + start
        if stop <= 0: stop = len(self.rules) + stop
        for r, rule in enumerate(self.rules.values()):
            if not (start <= r < stop): continue
            print(rule)

    def dump(self, fname):

        tokens = set([tok for stmt in self.statements.values() for tok in stmt.tokens])
        variables = set([var for rule in self.rules.values() for var in rule.variables])
        constants = tokens - variables

        # print(variables)
        # print(constants)
        # with open(fname, "w") as f: f.write(s)

@profile
def parse(fpath, max_rules=-1, last_rule=""):

    db = Database()

    # parser state
    in_comment = False # whether currently in comment
    current_tag = None # most recent tag (excluding comments)
    label = None # most recent label
    statement = None # most recent statement
    rule = None # most recent rule
    frames = [new_frame()] # stack of frames in current scope

    # dbg = False

    with open(fpath, "r") as f:
        for n, line in enumerate(f):

            if len(db.rules) == max_rules: break
            if rule is not None and rule.consequent.label == last_rule: break

            for token in line.strip().split():

                # if label == "df-alsc": dbg = True
                # if dbg:
                #     db.print(start=-2)
                #     print(f"token='{token}', label='{label}', tag='{current_tag}'")
                #     print(statement)
                #     input('..')    

                # skip comments
                if token == "$(": in_comment = True
                elif token == "$)": in_comment = False
                if in_comment: continue

                # update scope
                if token == "${": frames.append(new_frame())
                elif token == "$}": frames.pop()

                # initialize declarations
                elif token in ("$c", "$v", "$d"):
                    statement = Statement(label, token, [], [])

                # initialize labeled statements
                elif token in ("$f", "$e", "$a", "$p"):
                    assert label != None, \
                           f"line {n+1}: {token} not preceded by label"

                    statement = Statement(label, token, [], [])
                    db.statements[label] = statement

                # handle non-tag tokens
                elif token[0] != "$":

                    # update label
                    label = token

                    # update statements
                    if current_tag is not None:
                        if current_tag in "cvdfeap":
                            statement.tokens.append(token)
                        elif current_tag == "=":
                            statement.proof.append(token)

                # handle completed statements and rules
                elif token == "$.":

                    # update frame
                    if current_tag in "cv":
                        frames[-1][current_tag].extend(statement.tokens)
                    if current_tag == "d":
                        frames[-1][current_tag].append(sorted(statement.tokens))
                    if current_tag in "fe":
                        frames[-1][current_tag].append(statement)

                    # include placeholder "rules" for hypotheses
                    if current_tag in "fea=":
                        rule = Rule(statement, [], [], set(), set())

                    # attach scope to axioms and propositions
                    if current_tag in "a=":

                        # get all variables and essential hypotheses in current scope
                        for frame in frames:
                            rule.variables.update(frame["v"])
                            rule.essentials.extend(frame["e"])

                        # identify mandatory variables
                        tokens = set(statement.tokens)
                        for essential in rule.essentials:
                            tokens.update(essential.tokens)
                        mandatory = rule.variables & tokens

                        # save mandatory floating hypotheses
                        for frame in frames:
                            for floating in frame["f"]:
                                if floating.tokens[1] in mandatory:
                                    rule.floatings.append(floating)

                        # get disjoint variable pairs
                        for frame in frames:
                            for disjoint in frame["d"]:
                                rule.disjoint.update(it.combinations(disjoint, 2))

                    # add completed statements and finalized rules to database
                    if current_tag in "fea=":
                        rule.finalize()
                        db.statements[statement.label] = statement
                        db.rules[statement.label] = rule

                # update current tag
                if token[0] == "$" and token[1] not in "()": current_tag = token[1]
                if current_tag in ("$.", "$}"): current_tag = None

    return db

if __name__ == "__main__":

    import os

    fpath = os.path.join(os.environ["HOME"], "metamath", "set.mm")
    # fpath = 'badparse.mm'
    # fpath = "p2.mm"

    db = parse(fpath, max_rules=100)

    db.print(start=-6, stop=0)
    # # db.print()
    # print(f"{len(db.statements)} statements total, {len(db.rules)} rules total")

    # for (label, stmt) in db.statements.items():
    #     print(label, stmt)
    
    # print(db.rules['df-alsc'])
    # print(db.rules['alsconv'])

